from .unionfind import UnionFind
from ..reeb.lowerstar import LowerStar
import numpy as np


def is_face(sigma, tau):
    """
    Check if tau is a face of sigma

    Args:
        sigma: A simplicial complex (a list of simplices).
        tau: A simplex (a list of vertices).

    Returns:
        bool: True if tau is a face of sigma, False otherwise.
    """
    return set(tau).issubset(set(sigma))


def get_levelset_components(L):
    """
    Given a list of simplices L representing a level set, compute the connected components. This function is really only helpful inside of computeReeb.

    Args:
        L: A list of simplices (each simplex is a list of vertices).

    Returns:
        dict: A dictionary where keys are representative simplices and values are lists of simplices in the same connected component.
    """

    UF = UnionFind(range(len(L)))
    for i, simplex1 in enumerate(L):
        for j, simplex2 in enumerate(L):
            if i < j:
                # Check if they share a vertex
                if is_face(simplex1, simplex2) or is_face(simplex2, simplex1):
                    UF.union(i, j)

    # Replace indices with simplices
    components_index = UF.components_dict()
    components = {}
    for key in components_index:
        components[tuple(L[key])] = [L[i] for i in components_index[key]]

    return components


def computeReeb(K: LowerStar, verbose=False):
    """Computes the Reeb graph of a Lower Star Simplicial Complex K.

    Args:
        K (LowerStar): A Lower Star Simplicial Complex with assigned filtration values.
        verbose (boolean): Make it True if you want lots of printouts.

    Returns:
        ReebGraph: The computed Reeb graph.

    Example:
        >>> from cereeberus.reeb.LowerStar import LowerStar
        >>> K = LowerStar()
        >>> K.insert([0, 1, 2])
        >>> K.insert([1, 3])
        >>> K.insert([2,3])
        >>> K.assign_filtration([0], 0.0)
        >>> K.assign_filtration([1], 3.0)
        >>> K.assign_filtration([2], 5.0)
        >>> K.assign_filtration([3], 7)
        >>> R = computeReeb(K)
        >>> R.draw()
    """
    from ..reeb.reebgraph import ReebGraph

    funcVals = [(i, K.filtration([i])) for i in K.iter_vertices()]
    funcVals.sort(key=lambda x: x[1])  # Sort by filtration value

    R = ReebGraph()

    currentLevelSet = []
    components = {}
    half_edge_index = 0

    # This will keep track of the components represented by every vertex in the graph so far.
    # It will be vertName: connected_component (given as a list of lists) represented by that vertex
    vert_to_component = {}

    edges_at_prev_level = []

    for i, (vert, filt) in enumerate(funcVals):
        if verbose:
            print(f"\n---\n Processing {vert} at func val {filt:.2f}")
        now_min = filt
        now_max = funcVals[i + 1][1] if i + 1 < len(funcVals) else np.inf
        star = K.get_star([vert])
        lower_star = [s[0] for s in star if s[1] <= filt and len(s[0]) > 1]
        upper_star = [s[0] for s in star if s[1] > filt and len(s[0]) > 1]

        if verbose:
            print(f"  Lower star simplices: {lower_star}")
            print(f"  Upper star simplices: {upper_star}")

        # ----
        # Update the levelset list
        # ----

        for s in lower_star:
            # Remove from current level set
            if s in currentLevelSet:
                currentLevelSet.remove(s)

        currentLevelSet.append([vert])  # Add the vertex itself to the level set
        components_at_vertex = get_levelset_components(currentLevelSet)

        if verbose:
            print(f"  Current level set simplices: {currentLevelSet}")
            print(f"  Level set components at vertex {vert} (func val {filt:.2f}):")
            for comp in components_at_vertex.values():
                print(f"    Component: {comp}")

        verts_at_level = []
        for rep, comp in components_at_vertex.items():
            # Add a vertex for each component in this levelset
            nextNodeName = R.get_next_vert_name()
            R.add_node(nextNodeName, now_min)
            vert_to_component[nextNodeName] = (
                comp  # Store the component represented by this vertex
            )
            verts_at_level.append(nextNodeName)

            # Check if any simplex in vertex component is a subset of any of simplices in a previous edge's component
            for e in edges_at_prev_level:
                prev_comp = vert_to_component[e]
                if any(
                    [
                        is_face(prev_simp, simp)
                        for simp in comp
                        for prev_simp in prev_comp
                    ]
                ):
                    R.add_edge(e, nextNodeName)

        # ----
        # Add the edge vertices for after the vertex is passed
        # ----

        # Remove the vertex from the level set
        if [vert] in currentLevelSet:
            currentLevelSet.remove([vert])

        # Add the upper star to the current level set
        for s in upper_star:
            if s not in currentLevelSet:
                currentLevelSet.append(s)

        components = get_levelset_components(currentLevelSet)
        if verbose:
            print(f"\n  Updated current level set simplices: {currentLevelSet}")
            print(f"  Level set components after vertex {vert} (func val {filt:.2f}):")
            for comp in components.values():
                print(f"    Component: {comp}")
        # ----
        # Set up a vertex in the Reeb graph for each connected component
        # These will represent edges
        # These are at height (now_min + now_max)/2
        # ----
        edges_at_prev_level = []
        for comp in components.values():
            # Create a new vertex in the Reeb graph
            e_name = "e_" + str(half_edge_index)
            R.add_node(e_name, (now_min + now_max) / 2)
            vert_to_component[e_name] = (
                comp  # Store the component represented by this half edge top
            )
            half_edge_index += 1
            edges_at_prev_level.append(e_name)

            # Now connect to the vertices at this level
            for v in verts_at_level:

                # Get the component represented by vertex v
                prev_comp = vert_to_component[v]

                if any(
                    [
                        is_face(simp, prev_simp)
                        for simp in comp
                        for prev_simp in prev_comp
                    ]
                ):
                    R.add_edge(v, e_name)

    return R



import numpy as np
import matplotlib.pyplot as plt
from scipy.linalg import block_diag


class LabeledMatrix:
    """
    A class to store a matrix with row and column labels. The matrix is stored as a numpy array, and the row and columns are stored as lists. For example, we could store the 2 x 3 matrix below as a LabeledMatrix.

    +-------+---+---+---+
    |       | a | b | c |
    +=======+===+===+===+
    | **u** | 0 | 1 | 0 |
    +-------+---+---+---+
    | **v** | 1 | 0 | 0 |
    +-------+---+---+---+

    Where the matrix would be initialized by

    .. code-block:: python

        Mat = LabeledMatrix([[0, 1, 0], [1, 0, 0]], ['u', 'v'], ['a', 'b', 'c'])

    This class supports matrix multiplication, addition, and subtraction, as well as transposition and getting the maximum (absolute) value of the matrix. Note that calling for an entry in the matrix is done by the labels of the row and column, not the indices. For example, to get the entry in the first row and second column from the example above, you could call either of the following options.

    .. code-block:: python

        Mat['u', 'b']
        Mat.array[0, 1]

    Visualization can be done with the draw method, which will draw the matrix with the row and column labels.

    .. code-block:: python

        Mat.draw(colorbar = True)

    .. figure:: ../../images/labeled_matrix_ex_Mat.png
        :figwidth: 400px

    """

    def __init__(self, array=None, rows=None, cols=None, map_dict=None):
        """
        Initialize the LabeledMatrix with a numpy array, row labels, and column labels.

        Parameters:
            array (array-like): The matrix data.
            rows (list): The labels for the rows.
            cols (list): The labels for the columns.
        """
        self.rows = rows
        self.cols = cols

        if array is not None:
            self.array = np.array(array)
        elif map_dict is not None:
            self.array = np.zeros((len(rows), len(cols)))
            for c in cols:
                r = map_dict[c]
                # print (f"r = {r}, c = {c}")
                # print(f"rows: {rows}, cols: {cols}")
                self.array[rows.index(r), cols.index(c)] = 1
        elif rows is not None and cols is not None:
            self.array = np.zeros((len(rows), len(cols)))
        else:
            self.array = None

    def __matmul__(self, other):
        """
        Multiply two LabeledMatrix objects using the @ operator.

        Parameters:
            other (LabeledMatrix): The other LabeledMatrix to multiply with.

        Returns:
            LabeledMatrix: The result of the matrix multiplication.

        Raises:
            ValueError: If the other object is not a LabeledMatrix.
            ValueError: If the matrix sizes are incompatable, or if they are but the row labels of the first matrix do not match the column labels of the second (including ordering).
        """

        if isinstance(other, LabeledBlockMatrix):
            other = other.to_labeled_matrix()
        elif not isinstance(other, LabeledMatrix):
            raise ValueError(
                "Can only multiply with another LabeledMatrix or LabeledBlockMatrix"
            )

        if self.cols != other.rows:
            raise ValueError(
                f"Matrix labels do not match.\nCols of first matrix: {self.cols}\nRows of second matrix: {other.rows}"
            )

        # Perform matrix multiplication
        result_array = self.array @ other.array

        # The row labels of the result are the row labels of the first matrix
        result_rows = self.rows

        # The column labels of the result are the column labels of the second matrix
        result_cols = other.cols

        return LabeledMatrix(result_array, result_rows, result_cols)

    def __add__(self, other):
        """
        Add two LabeledMatrix objects using the + operator.

        Parameters:
            other (LabeledMatrix): The other LabeledMatrix to add.

        Returns:
            LabeledMatrix: The result of the matrix addition.

        Raises:
            ValueError: If the other object is not a LabeledMatrix.
        """
        if not isinstance(other, LabeledMatrix):
            raise ValueError("Can only add with another LabeledMatrix")

        if not (self.rows == other.rows and self.cols == other.cols):
            raise ValueError("Cannot add matrices with different row or column labels.")

        # Perform matrix addition
        result_array = self.array + other.array

        return LabeledMatrix(result_array, self.rows, self.cols)

    def __sub__(self, other):
        """
        Subtract two LabeledMatrix objects using the - operator.

        Parameters:
            other (LabeledMatrix): The other LabeledMatrix to subtract.

        Returns:
            LabeledMatrix: The result of the matrix subtraction.

        Raises:
            ValueError: If the other object is not a LabeledMatrix.
        """
        if not isinstance(other, LabeledMatrix):
            raise ValueError("Can only subtract with another LabeledMatrix")

        # Perform matrix subtraction
        result_array = self.array - other.array

        return LabeledMatrix(result_array, self.rows, self.cols)

    def __getitem__(self, key):
        """
        Get an item from the matrix, where ``key = [row_key, col_key]`` are the **labels** of the row and columns respectively.

        Parameters:
            key: The keys for the row and column to get from the matrix.

        Returns:
            The item from the matrix.
        """
        row_key, col_key = key
        i = self.rows.index(row_key)
        j = self.cols.index(col_key)
        return self.array[i, j]

    def __setitem__(self, key, value):
        """
        Set an item in the matrix, where ``key = [row_key, col_key]`` are the **labels** of the row and columns respectively.

        Parameters:
            key: The row and column keys to set in the matrix.
            value: The value to set.
        """
        row_key, col_key = key
        i = self.rows.index(row_key)
        j = self.cols.index(col_key)
        self.array[i, j] = value

    def get_array(self):
        """
        Get the array of the matrix. Works on the big matrix as well as the blocks.

        Returns:
            np.array: The array of the matrix.
        """
        return self.array

    def max(self):
        """
        Get the maximum value of the matrix.

        Returns:
            float: The maximum value of the matrix.
        """
        try:
            return np.max(self.array)
        except:
            return np.nan

    def absmax(self):
        """
        Get the maximum absolute value of the matrix.

        Returns:
            float: The maximum absolute value of the matrix.
        """
        try:
            return np.max(np.abs(self.array))
        except:
            return np.nan

    def min(self):
        """
        Get the minimum value of the matrix.

        Returns:
            float: The minimum value of the matrix.
        """
        try:
            return np.min(self.array)
        except:
            return np.nan

    def shape(self):
        """
        Return the shape of the matrix.

        Returns:
            tuple: The shape of the matrix (rows, columns).
        """
        try:
            return self.array.shape
        except:
            return (0, 0)

    def size(self):
        """
        Return the size of the matrix.

        Returns:
            int: The size of the matrix.
        """
        try:
            return self.array.size
        except:
            return 0

    def __repr__(self):
        """
        Return a string representation of the LabeledMatrix.

        Returns:
            str: A string representation of the LabeledMatrix.
        """
        return f"LabeledMatrix(\narray=\n{self.array}, \nrows={self.rows}, \ncols={self.cols})"

    def T(self):
        """
        Transpose the matrix.

        Returns:
            LabeledMatrix: The labeled transposed matrix.
        """
        return LabeledMatrix(self.array.T, self.cols, self.rows)

    def to_labeled_block(self, row_val_to_verts, col_val_to_verts):
        """
        Turns this matrix back into a block dictionary.

        Note this assumes that the row and column labels are already in order from the block dictionary.

        Parameters:
            row_val_to_verts : dict
                A dictionary from function values to a list of row objects.
            col_val_to_verts : dict
                A dictionary from function values to a list of column objects.

        Returns:
            LabeledBlockMatrix
                A block matrix with the same data as the labeled matrix.
        """

        all_keys = list(row_val_to_verts.keys()) | set(col_val_to_verts.keys())
        all_keys.sort()

        blocks_out = LabeledBlockMatrix()

        curr_row = 0
        curr_col = 0

        for i in all_keys:
            try:
                rows = row_val_to_verts[i]
                next_row = curr_row + len(rows)
            except KeyError:
                rows = []
                next_row = curr_row

            try:
                cols = col_val_to_verts[i]
                next_col = curr_col + len(col_val_to_verts[i])
            except KeyError:
                cols = []
                next_col = curr_col

            A = self.array[curr_row:next_row, curr_col:next_col]
            blocks_out.blocks[i] = LabeledMatrix(A, rows, cols)

            curr_row = next_row
            curr_col = next_col

        return blocks_out

    def col_sum(self):
        """
        Returns the sum of each column in the matrix.
        """
        try:
            return self.array.sum(axis=0)
        except:
            return []

    def is_col_sum_1(self):
        """
        Check if the sum of each column is 1.

        Returns:
            bool: True if the sum of each column is 1, False otherwise.
        """
        try:
            return np.all(self.array.sum(axis=0) == 1)
        except:
            return False

    def draw(self, ax=None, colorbar=False, **kwargs):
        """
        Draw the matrix with row and column labels.

        Parameters:
            ax (matplotlib.axes.Axes): The axes to draw the matrix on. If None, the current axes will be used.
            colorbar (bool): Whether to draw a colorbar.
            **kwargs: Additional keyword arguments to pass to ax.matshow.
        """

        if ax is None:
            ax = plt.gca()

        im = ax.matshow(self.array, **kwargs)

        # Add vertices as the row and column labels

        if type(self.rows[0]) == tuple:
            row_labels = [str(row[0]) + ", " + str(row[1]) for row in self.rows]
            # print('Warning: Row labels are tuples. Only printing entries 0 and 1. ')
            # This is done because in multigraphs, edges are (u,v,count) and we only want to print u and v

        else:
            row_labels = self.rows

        if type(self.cols[0]) == tuple:
            col_labels = [str(col[0]) + ", " + str(col[1]) for col in self.cols]
            # print('Warning: Col labels are tuples. Only printing entries 0 and 1. ')
        else:
            col_labels = self.cols

        ax.set_xticks(range(len(self.cols)), col_labels, rotation=90)
        ax.set_yticks(range(len(self.rows)), row_labels)

        if colorbar:
            plt.colorbar(im, ax=ax)

        return ax

    def to_indicator(self):
        """
        Convert the matrix to an indicator matrix.

        Returns:
            LabeledMatrix: The indicator matrix.
        """
        return LabeledMatrix((self.array > 0).astype(int), self.rows, self.cols)


########################################################################################


class LabeledBlockMatrix:
    """
    A class to store a block matrix with row and column labels. This is built with the assumption that we have a function value for each object in the labels, and we initialize this with a dictionary for rows and columns that sends function value ``i`` to a list of row or column labels, respectively.

    For example, we could store the block matrix below as a ``LabeledBlockMatrix``. The outside rows give the function values of the vertices, and the inside rows and columns give the labels of the vertices.

    +--------+-------+---+---+---+---+---+
    |                |   1   |   2       |
    +                +---+---+---+---+---+
    |                | a | b | c | d | e |
    +========+=======+===+===+===+===+===+
    | **0**  | **u** | 0 | 0 | 0 | 0 | 0 |
    +--------+-------+---+---+---+---+---+
    | **1**  | **v** | 1 | 1 | 0 | 0 | 0 |
    +        +-------+---+---+---+---+---+
    |        | **w** | 0 | 0 | 0 | 0 | 0 |
    +--------+-------+---+---+---+---+---+
    | **2**  | **x** | 0 | 0 | 1 | 1 | 1 |
    +--------+-------+---+---+---+---+---+

    The matrix is stored as a dictionary of LabeledMatrix objects, where the keys are the function values of the vertices. For example, we could store the block matrix below as a ``LabeledBlockMatrix``

    .. code-block:: python

        cols_dict = {1: ['a','b'], 2: ['c','d','e']}
        rows_dict = {0: ['u'], 1: ['v','w'], 2: ['x']}
        map_dict = {'a': 'v', 'b':'v', 'c': 'x', 'd':'x', 'e':'x'}
        lbm = LabeledBlockMatrix(map_dict, rows_dict, cols_dict)

    Note that this can either be drawn with entries filled with either 0's or ``np.nan``.

    .. code-block:: python

        lbm.draw(colorbar = True)

    .. figure:: ../../images/labeled_matrix_ex_BlockMat_0.png
        :figwidth: 400px

    .. code-block:: python

        lbm.draw(filltype = 'nan', colorbar = True)

    .. figure:: ../../images/labeled_matrix_ex_BlockMat_nan.png
        :figwidth: 400px

    This class supports matrix multiplication, addition, and subtraction, as well as transposition and getting the maximum (absolute) value of the matrix. Note that calling for an item returns the i'th block as a ``LabeledMatrix``. So we could get an entry in the matrix by first calling the block number, and then the row and column labels.

    .. code-block:: python

            lbm[1]['v', 'b']


    """

    def __init__(
        self,
        map_dict=None,
        rows_dict=None,
        cols_dict=None,
        random_initialize=False,
        labled_matrix_dict=None,
        seed=None,
    ):
        """
        Initialize the LabeledBlockMatrix with LabeledMatrix objects for each integer.

        rows_dict and cols_dict are dictionaries from integers to lists of row labels and column labels respectively. The keys of the dictionaries are the function values of the vertices.

        If map_dict is provided, it is a dictionary from column objects to row objects. The matrix will be filled in with 1s where the column object maps to the row object. If map_dict is None, the matrix will be filled in with random 1s in each column if random_initialize is True. If none of these are provided, the matrix will be filled with zeros.


        Parameters:
            rows_dict : dict
                A dictionary from integers to lists of row labels.
            cols_dict : dict
                A dictionary from integers to lists of column labels.
            map_dict : dict
                A dictionary from column objects to row objects.
            random_initialize : bool
                Whether to randomly initialize the matrix.
            labeled_matrix_dict : bool
                Whether to initialize the block matrix from a dictionary of LabeledMatrix objects.
            seed : int
                The seed for the random number generator.
        """

        self.blocks = {}

        if (
            labled_matrix_dict is not None
        ):  # Initialize from a dictionary of LabeledMatrix objects
            # make sure all the other dicts are none
            if (
                rows_dict is not None
                or cols_dict is not None
                or map_dict is not None
                or random_initialize
            ):
                raise ValueError(
                    "Cannot initialize from a dictionary of LabeledMatrix objects and other dictionaries. random initialization must be False."
                )

            for i in labled_matrix_dict.keys():
                self.blocks[i] = labled_matrix_dict[i]

        if rows_dict is not None:
            all_keys = list(set(rows_dict.keys()) | set(cols_dict.keys()))
            all_keys.sort()

            for i in all_keys:
                # i is the function value of the vertices

                # Initialize the block and fill with empty matrices
                if i in rows_dict:
                    rows_ = rows_dict[i]
                else:
                    rows_ = []

                if i in cols_dict:
                    cols_ = cols_dict[i]
                else:
                    cols_ = []

                self.blocks[i] = LabeledMatrix(rows=rows_, cols=cols_)

                # If a map from column objects to row objects is provided, fill in the matrix
                if map_dict is not None:
                    if cols_ and rows_:
                        for c in cols_:
                            try:
                                r = map_dict[c]
                            except:
                                print("Error")
                                print(f"c: {c}")
                                print(f"map_dict: {map_dict}")
                                print(f"cols_: {cols_}")
                                print(f"rows_: {rows_}")
                                print(f"blocks[i]: {self.blocks[i]}")
                                ValueError("Column object {c} not in map_dict")
                            self.blocks[i][r, c] = 1
                elif random_initialize:
                    A = self.blocks[i].array
                    rng = np.random.default_rng(seed=seed)
                    col_1s = rng.integers(0, A.shape[0], size=A.shape[1])
                    A[col_1s, list(range(A.shape[1]))] = 1
                    self.blocks[i].array = A

    def __repr__(self):
        """
        Return a string representation of the LabeledBlockMatrix.

        Returns:
        str: A string representation of the LabeledBlockMatrix.
        """
        return f"LabeledBlockMatrix(matrices={self.blocks})"

    def __getitem__(self, key):
        """
        Get the i'th block matrix.

        Parameters:
            key: The key to get from the block matrix.

        Returns:
            The item from the block matrix.
        """
        return self.blocks[key]

    def __setitem__(self, key, value):
        """
        Set the i'th block matrix.

        Parameters:
            key: The key to set in the block matrix.
            value: The value to set.
        """
        self.blocks[key] = value

    def __matmul__(self, other):
        """
        Multiply two LabeledBlockMatrix objects using the @ operator.

        Parameters:
            other (LabeledBlockMatrix): The other LabeledBlockMatrix to multiply with.

        Returns:
            LabeledBlockMatrix: The result of the matrix multiplication.

        Raises:
            ValueError: If the other object is not a LabeledBlockMatrix.
        """
        if isinstance(other, LabeledBlockMatrix):

            # Perform matrix multiplication
            all_keys = list(set(self.blocks.keys()) | set(other.blocks.keys()))
            all_keys.sort()

            result = LabeledBlockMatrix()

            for i in all_keys:
                if i not in self.blocks:
                    result[i] = LabeledMatrix(cols=other.blocks[i].cols, rows=[])
                elif i not in other.blocks:
                    result[i] = LabeledMatrix(rows=self.blocks[i].rows, cols=[])
                else:
                    result[i] = self.blocks[i] @ other.blocks[i]

            return result

        elif isinstance(other, LabeledMatrix):
            return self.to_labeled_matrix() @ other

        else:
            raise ValueError(
                "Can only multiply with another LabeledBlockMatrix or LabeledMatrix"
            )

    def __add__(self, other):
        """
        Add two LabeledBlockMatrix objects using the + operator.

        Parameters:
            other (LabeledBlockMatrix): The other LabeledBlockMatrix to add.

        Returns:
            LabeledBlockMatrix: The result of the matrix addition.

        Raises:
            ValueError: If the other object is not a LabeledBlockMatrix.
        """
        if not isinstance(other, LabeledBlockMatrix):
            raise ValueError("Can only add with another LabeledBlockMatrix")

        # Perform matrix addition
        all_keys = set(self.blocks.keys()) | set(other.blocks.keys())

        result = LabeledBlockMatrix()

        for i in all_keys:
            if i not in self.blocks:
                result[i] = other.blocks[i]
            elif i not in other.blocks:
                result[i] = self.blocks[i]
            else:
                result[i] = self.blocks[i] + other.blocks[i]

        return result

    def __sub__(self, other):
        """
        Subtract two LabeledBlockMatrix objects using the - operator.

        Parameters:
            other (LabeledBlockMatrix): The other LabeledBlockMatrix to subtract.

        Returns:
            LabeledBlockMatrix: The result of the matrix subtraction.

        Raises:
            ValueError: If the other object is not a LabeledBlockMatrix.
        """
        if not isinstance(other, LabeledBlockMatrix):
            raise ValueError("Can only subtract with another LabeledBlockMatrix")

        # Perform matrix subtraction
        all_keys = set(self.blocks.keys()) | set(other.blocks.keys())

        result = LabeledBlockMatrix()

        for i in all_keys:
            if i not in self.blocks:
                result[i] = other.blocks[i]
            elif i not in other.blocks:
                result[i] = self.blocks[i]
            else:
                result[i] = self.blocks[i] - other.blocks[i]

        return result

    def shape(self):
        """
        Return the shape of the block matrix.

        Returns:
            tuple: The shape of the block matrix.
        """
        try:
            return (
                sum([self.blocks[i].shape()[0] for i in self.blocks.keys()]),
                sum([self.blocks[i].shape()[1] for i in self.blocks.keys()]),
            )
        except:
            return (0, 0)

    def T(self):
        """
        Transpose the block matrix.

        Returns:
            LabeledBlockMatrix: The labeled transposed block matrix.
        """
        result = LabeledBlockMatrix()

        for i in self.blocks.keys():
            result[i] = self.blocks[i].T()

        return result

    def get_array(self, block_index):
        """
        Get the array of the block matrix at the given block index.

        Parameters:
            block_index (int): The block index to get the array from.

        Returns:
            np.array: The array of the block matrix at the given block index.
        """
        return self.blocks[block_index].array

    def max(self):
        """
        Get the maximum value of the block matrix.

        Returns:
            float: The maximum value of the block matrix.
        """
        try:
            return np.nanmax([self.blocks[i].max() for i in self.blocks.keys()])
        except:
            return np.nan

    def absmax(self):
        """
        Get the maximum absolute value of the block matrix.

        Returns:
            float: The maximum absolute value of the block matrix.
        """
        try:
            return np.nanmax([self.blocks[i].absmax() for i in self.blocks.keys()])
        except:
            return np.nan

    def min(self):
        """
        Get the minimum value of the block matrix.

        Returns:
            float: The minimum value of the block matrix.
        """
        try:
            return np.nanmin([self.blocks[i].min() for i in self.blocks.keys()])
        except:
            return np.nan

    def get_all_block_indices(self):
        """
        Get all the block indices in order.

        Returns:
            list: The block indices.
        """
        L = list(self.blocks.keys())
        L.sort()
        return L

    def min_block_index(self):
        """
        Get the minimum block index.

        Returns:
            int: The minimum block index.
        """
        return min(self.blocks.keys())

    def max_block_index(self):
        """
        Get the maximum block index.

        Returns:
            int: The maximum block index.
        """
        return max(self.blocks.keys())

    def get_all_rows(self):
        """
        Get all the row labels of the block matrix in order.

        Returns:
            list: The row labels.
        """
        all_keys = self.get_all_block_indices()

        rows = [self.blocks[i].rows for i in all_keys]

        # flatten the rows list

        return [label for sublist in rows for label in sublist]

    def get_all_cols(self):
        """
        Get all the col labels of the block matrix in order .

        Returns:
            list: The col labels.
        """
        all_keys = self.get_all_block_indices()

        cols = [self.blocks[i].cols for i in all_keys]

        return [label for sublist in cols for label in sublist]

    def get_all_labels(self):
        """
        Get all the row and column of the block matrix in function value order.

        Returns:
            tuple: The row and column labels.
        """
        return self.get_all_rows(), self.get_all_cols()

    def block_diag_NaN(self, *arrs):
        """
        Create a single matrix from provided arrays on the blocks. Modified from `scipy.linalg.block_diag` to put np.nan in the off-diagonal blocks instead of 0's.

        Parameters:
            *arrs : array-like
                The arrays to put on the diagonal.

        Returns:
            np.array: The block diagonal matrix.

        """
        if arrs == ():
            arrs = ([],)
        arrs = [np.atleast_2d(a) for a in arrs]

        bad_args = [k for k in range(len(arrs)) if arrs[k].ndim > 2]
        if bad_args:
            raise ValueError(
                "arguments in the following positions have dimension "
                "greater than 2: %s" % bad_args
            )

        shapes = np.array([a.shape for a in arrs])
        # out_dtype = np.result_type(*[arr.dtype for arr in arrs])
        # out = np.empty(np.sum(shapes, axis=0), dtype=out_dtype)
        out = np.empty(np.sum(shapes, axis=0))
        out.fill(np.nan)

        r, c = 0, 0
        for i, (rr, cc) in enumerate(shapes):
            out[r : r + rr, c : c + cc] = arrs[i]
            r += rr
            c += cc
        return out

    def to_labeled_matrix(self, filltype="zero"):
        """
        Convert to a single block diagonal matrix.

        returns:
            LabeledMatrix :
                A labeled matrix with the same data as the block matrix.
        """

        all_keys = self.get_all_block_indices()
        arrays = [self.blocks[i].array for i in all_keys]

        if filltype == "zero":
            BigMatrix = block_diag(*arrays)
        elif filltype == "nan":
            BigMatrix = self.block_diag_NaN(*arrays)

        rows = self.get_all_rows()
        cols = self.get_all_cols()

        return LabeledMatrix(BigMatrix, rows, cols)

    def to_indicator(self):
        """
        Convert the block matrix to an indicator matrix.

        Returns:
            LabeledBlockMatrix: The indicator matrix.
        """
        result = LabeledBlockMatrix()

        for i in self.blocks.keys():
            result[i] = self.blocks[i].to_indicator()

        return result

    def to_shifted_blocks(self, shift):
        """
        Shift the blocks in the block matrix.

        Parameters:
            shift (int): The amount to shift the blocks.

        Returns:
            LabeledBlockMatrix: The block matrix with shifted blocks.
        """
        result = LabeledBlockMatrix()

        for i in self.blocks.keys():
            result[i + shift] = self.blocks[i]

        return result

    def col_sum(self):
        """
        Returns the sum of each column in the block matrix.

        Returns:
            np.array: The sum of each column in the block matrix.
        """
        try:
            sorted_keys = list(self.blocks.keys())
            sorted_keys.sort()
            return np.concatenate([self.blocks[i].col_sum() for i in sorted_keys])
        except:
            return []

    def check_column_sum(self, verbose=False):
        """
        Check that the sum of each column is 1.

        Parameters:
            verbose (bool):
                Prints information on which matrices have columns that do not sum to 1 if True. The default is False.

        Returns:
            bool
                True if the columns sum to 1, False otherwise.
        """

        for i in self.blocks.keys():
            if not self.blocks[i].is_col_sum_1():
                if verbose:
                    print(f"Column sum is not 1 for block {i}")
                return False

        return True

    def draw(self, ax=None, colorbar=False, filltype="zero", **kwargs):
        """
        Draw the block matrix with row and column labels.

        Parameters:
            ax (matplotlib.axes.Axes): The axes to draw the matrix on. If None, the current axes will be used.
            colorbar (bool): Whether to draw a colorbar.
            filltype (str): Either ``zeros`` or ``nan``. If ``zeros``, the off-diagonal blocks will be filled with zeros prior to drawing. If ``nan``, the off-diagonal blocks will be filled with ``np.nan``, resulting in them showing up white.
            **kwargs: Additional keyword arguments to pass to ``ax.matshow``.
        """

        self.to_labeled_matrix(filltype=filltype).draw(
            ax=ax, colorbar=colorbar, **kwargs
        )

        return ax

    def get_rows_dict(self):
        """
        Get the rows dictionary.

        Returns:
            dict: The rows dictionary.
        """
        result = {}
        for i in self.blocks.keys():
            result[i] = self.blocks[i].rows

        return result

    def get_cols_dict(self):
        """
        Get the columns dictionary.

        Returns:
            dict: The columns dictionary.
        """
        result = {}
        for i in self.blocks.keys():
            result[i] = self.blocks[i].cols

        return result



from ..reeb.mapper import MapperGraph


# Interprets the lensfunction as a python function, does it, then returns the new location of each point for every point
def __runlensfunction(lensfunction, pointcloud):
    lensfunctionoutput = []
    if callable(lensfunction):
        for val in range(len(pointcloud)):
            lensfunctionoutput.append(
                [lensfunction(pointcloud[val]), tuple(pointcloud[val])]
            )
        return lensfunctionoutput

    if len(lensfunction) != len(pointcloud):
        raise ValueError(
            "If the lens function is given as a list of numbers, it must have the same length as the number of points in the point cloud."
        )

    for val in range(len(pointcloud)):
        lensfunctionoutput.append([lensfunction[val], tuple(pointcloud[val])])

    return lensfunctionoutput


# Creates a list of covers, together with any points inside the cover, ie, [[cover, point1, point2], [cover, point3]]
# Also removes any covers that have no points inside of them
# this would probably be better done as a struct, containing a covering set, the covering set's position in the cover, and an array of points
def __createcoveringsets(points, cover):
    # adds location information to cover
    for val0 in range(len(cover)):
        cover[val0] = (cover[val0][0], cover[val0][1], val0)
    # creates the list
    coveringsets = []
    for val1 in range(len(cover)):
        coveringsets.append([cover[val1]])
        for val2 in range(len(points)):
            if points[val2][0] >= cover[val1][0] and points[val2][0] <= cover[val1][1]:
                coveringsets[val1].append(points[val2])
    # removes unused covers
    position = 0
    while position < len(coveringsets):
        if len(coveringsets[position]) == 1:
            coveringsets.pop(position)
        else:
            position += 1
    # print("Covering Sets Output: ")
    # print(coveringsets)
    return coveringsets


# cluster the points using a number of existing clustering algorithms
def __cluster(coveringsets, clusteralgorithm):
    # trivial clustering
    if clusteralgorithm == "trivial":
        finished_cluster = list()
        cluster = list()
        for val1 in range(len(coveringsets)):
            cluster.append(coveringsets[val1][0][2])
            for val2 in range(1, len(coveringsets[val1])):
                cluster.append(
                    (coveringsets[val1][val2][1][0], coveringsets[val1][val2][1][1])
                )
            finished_cluster.append(
                cluster[:]
            )  # Works like this to avoid passing by reference in python lists
            cluster.clear()
        # print("Clustering Output: ")
        # print(finished_cluster)
        return finished_cluster
    # execute sklearn clusterings
    elif callable(clusteralgorithm):
        finished_cluster = list()
        coverpointcloud = list()
        cluster = list()
        for val1 in range(len(coveringsets)):
            # alters data to fit with sklearn clustering algorithms
            for val2 in range(1, len(coveringsets[val1])):
                coverpointcloud.append(
                    (coveringsets[val1][val2][1][0], coveringsets[val1][val2][1][1])
                )
            # does clustering algorithm
            cluster_out = clusteralgorithm(coverpointcloud)
            # puts points into list
            for val2 in range(max(cluster_out.labels_) + 1):
                cluster.append(
                    [coveringsets[val1][0][2]]
                )  # The position of the covering set in the cover (preserved for distance purposes)
                for val3 in range(len(cluster_out.labels_)):
                    if cluster_out.labels_[val3] == val2:
                        cluster[val2].append(coverpointcloud[val3])
                finished_cluster.append(cluster[val2])
            coverpointcloud.clear()
            cluster.clear()
        # print("Clustering Output: ")
        # print(finished_cluster)
        return finished_cluster
    else:
        print("input not valid")
        return list()


# Adds edges between the cluster that share points
def __addedges(clusterpoints):
    outputgraph = MapperGraph()
    val2 = 0
    for val1 in range(len(clusterpoints)):
        outputgraph.add_node(val1, clusterpoints[val1][0])
        while val2 < val1:
            if clusterpoints[val1][0] != clusterpoints[val2][0]:
                if len(set(clusterpoints[val1]) & set(clusterpoints[val2])) > 0:
                    outputgraph.add_edge(val1, val2)
            val2 += 1
        val2 = 0
    # print("Final Output: ")
    # print(outputgraph)
    return outputgraph


# Does the Mapper Algorithm in order
def computeMapper(pointcloud, lensfunction, cover, clusteralgorithm):
    """
    Computes the mapper graph of an input function. 
    
    The point cloud should be given as a list of tuples or as a numpy array. 
    
    The lens function should be given as either a list of numbers with the same length as the number of points; or as a callable function where :math:`f(point) = \text{value}` so long as the function can be determined from the coordinate values of the point.    
    
    The cover should be given as a list of intervals. This can be done, for example, using the 'cereeberus.cover' function in this module, which takes in a minimum, maximum, number of covers, and percentage of overlap to create a cover. 
    
    The clustering algorithm should be given as a callable that takes in a point cloud and outputs cluster labels (for example, `sklearn.cluster.DBSCAN(min_samples=2,eps=0.3).fit`).

    Parameters:
        A pointcloud (as a list of tuples or as a numpy array)
        A lens function (as a callable or a list of numbers)
        A cover (as a list of intervals)
        A clustering algorithm (as a callable)

    Returns:
        A `MapperGraph` object representing the mapper graph of the input data and lens function.
    """
    
    lensfunctionoutput = __runlensfunction(lensfunction, pointcloud)                           
    coveringsets = __createcoveringsets(lensfunctionoutput, cover)
    clusterpoints = __cluster(coveringsets, clusteralgorithm)
    outputgraph = __addedges(clusterpoints)
    
    return outputgraph


# function to create covers
# cover(min, max, #covers, %overlap)
def cover(min=-1, max=1, numcovers=10, percentoverlap=0.5):
    """
    Creates a cover to be used for inputs in the computeMapper function

    Parameters:
        min: the minimum for the range of the covering sets
        max: the maximum for the range of the covering sets
        numcovers: number of covers to create
        percentoverlap: percentage (from 0 to 1) of overlap between covers

    Returns:
        An array of intervals
    """
    output = []
    val = 0
    coversize = (max - min) / numcovers * (1 + (percentoverlap))
    while val < numcovers:
        center = (min * (numcovers - (val + 0.5)) + max * (val + 0.5)) / numcovers
        output.append(((-0.5 * coversize) + center, (0.5 * coversize) + center))
        val += 1
    return output



import networkx as nx
from ..draw import draw
import matplotlib.pyplot as plt
import numpy as np


class ReebGraph(nx.MultiDiGraph):
    """
    A Reeb graph stored as a networkx ``MultiDiGraph``. The function values are stored as a dictionary. The directedness of the edges follows the convention that the edge goes from the lower function value to the higher function value node.

    """

    def __init__(self, G=None, f={}, seed=None, verbose=False):
        """Initializes a Reeb graph object.

        Parameters:
            G : nx.graph, optional. If not None, a graph to initialize the Reeb graph.
            f : dict, optional. If not an empty dictionary, a dictionary of function values associated with the graph nodes.
            seed : int, optional. If not None, a seed to pass to the spring layout.
            verbose: bool, optional. If True, will print out additional information during initialization.
        """

        super().__init__()

        if G is None:
            self.f = {}
        elif f != {}:
            self.f = f
        else:
            # Some example graphs still have the function values stored as 'fx'
            self.f = nx.get_node_attributes(G, "fx")

        if verbose:
            print(f"Function values: {self.f}")

        if G is not None:
            if verbose:
                print(f"Nodes: {G.nodes}")
            for v in G.nodes:
                self.add_node(v, self.f[v], reset_pos=False)

            # self.f = nx.get_node_attributes(G, 'f')

            for e in G.edges:
                e = e[:2]
                if verbose:
                    print(f"Edge: {e}")
                # Multigraph means edges are stored as a triple (u,v,n) where n is the number of the edge.
                # This is an old note, maybe not true anymore?
                edge = (e[0], e[1])
                self.add_edge(*edge, reset_pos=False)

        self.set_pos_from_f(seed=seed, verbose=verbose)

    def __str__(self):
        return f"ReebGraph with {len(self.nodes)} nodes and {len(self.edges)} edges."

    def summary(self):
        """Summary of the Reeb graph.

        Returns:
            dict
                A dictionary with the number of nodes and edges in the Reeb graph.
        """
        return {"nodes": len(self.nodes), "edges": len(self.edges)}

    # -----------------#
    # Methods for getting info about the Reeb graph and its nodes
    # -----------------#

    def get_function_values(self):
        """
        Get the function values of the nodes in the Reeb graph.

        Returns:
            list
                A sorted list of function values.
        """
        L = [self.f[v] for v in self.nodes]
        L.sort()
        return L

    def min_f(self):
        """
        Get the minimum function value in the Reeb graph.

        Returns:
            float
                The minimum function value.
        """
        return min(self.f.values())

    def max_f(self):
        """
        Get the maximum function value in the Reeb graph.

        Returns:
            float
                The maximum function value.
        """
        return max(self.f.values())

    def up_degree(self, node):
        """Get the up degree of a node.

        Parameters:
            node : int. The node to get the up degree of.

        Returns:
            int
                The up degree of the node.
        """
        return self.out_degree(node)

    def down_degree(self, node):
        """
        Get the down degree of a node.

        Parameters:
            node : int
                The node to get the down degree of.

        Returns:
            int
                The down degree of the node.
        """
        return self.in_degree(node)

    def number_connected_components(self):
        """
        Get the number of connected components in the Reeb graph.

        Returns:
            int
                The number of connected components in the Reeb graph.
        """
        return nx.number_connected_components(self.to_undirected())

    def func_to_vertex_dict(self):
        """
        Get a dictionary mapping function values to all vertices at that height.

        Returns:
            dict
                A dictionary mapping function values to vertices.
        """
        f_to_v = {}
        for v in self.nodes:
            f = self.f[v]
            if f in f_to_v:
                f_to_v[f].append(v)
            else:
                f_to_v[f] = [v]
        return f_to_v

    def sorted_vertices(self):
        """
        Get a list of vertices sorted by function value. Same order as passed by the func_to_vertex_dict method.

        Returns:
            list
                A list of vertices sorted by function value.
        """

        func_dict = self.func_to_vertex_dict()
        keys = list(func_dict.keys())
        keys.sort()
        vertices = [func_dict[k] for k in keys]
        vertices = [v for sublist in vertices for v in sublist]

        return vertices

    def func_to_edge_dict(self):
        """
        Get a dictionary mapping function values to all edges with lower endpoint at that height.

        Returns:
            dict
                A dictionary mapping function values to edges.
        """
        f_to_e = {}
        for e in self.edges:
            f = self.f[e[0]]
            if f in f_to_e:
                f_to_e[f].append(e)
            else:
                f_to_e[f] = [e]
        return f_to_e

    def sorted_edges(self):
        """
        Get a list of edges sorted by function value. Same order as passed by the func_to_edge_dict method.

        Returns:
            list
                A list of edges sorted by function value.
        """
        func_dict = self.func_to_edge_dict()
        keys = list(func_dict.keys())
        keys.sort()
        edges = [func_dict[k] for k in keys]
        edges = [item for row in edges for item in row]

        return edges

    def relabel_nodes(self, mapping):
        """
        Relabel the nodes of the Reeb graph using a mapping. The mapping should be a dictionary with keys as old node names and values as new node names.

        Parameters:
            mapping : dict. A dictionary mapping old node names to new node names.
        """

        # Store copies of the old stuff
        nodes_old = list(self.nodes()).copy()
        edges_old = list(self.edges(keys=True)).copy()
        f_old = self.f.copy()

        # Delete all nodes and edges from the graph
        for n in nodes_old:
            self.remove_node(n)

        # Add the nodes and edges back in with the new names
        for v_old in nodes_old:
            self.add_node(mapping[v_old], f_old[v_old])

        for e in edges_old:
            self.add_edge(mapping[e[0]], mapping[e[1]], e[2])

    def inv_image(self, f_val):
        """
        Get the set of vertices and/or edges at a given function value. This will return $v$ for which $f(v) = $`f_val`, as well as edges $e = (u,v)$ for which $f(u) < $`f_val` and $f(v) > $`f_val`.

        Parameters:
            f_val : float. The function value to get the objects for.

        Returns:
            tuple : (set, set)
                The set of vertices and the set of edges at the function value `f_val`.
        """
        verts = {v for v in self.nodes if self.f[v] == f_val}
        edges = {e for e in self.edges if self.f[e[0]] < f_val and self.f[e[1]] > f_val}
        return verts, edges

    # -----------------------------------------------#
    # Methods for getting distances in the Reeb graph
    # -----------------------------------------------#

    def thickening_distance(self, u, v):
        """
        Get the thickening distance between two vertices in the Reeb graph. This is the amount of thickening needed before the two vertices map to the same connected component. Note that u and v need to be at the same function value, so f(u) = f(v).

        Parameters:
            u, v : int. The vertices to get the thickening distance between.

        Returns:
            int
                The thickening distance between the two vertices.
        """
        if self.f[u] != self.f[v]:
            raise ValueError(
                f"Vertices {u} and {v} are not at the same function value."
            )

        a = self.f[u]

        # Get function values
        all_f = self.get_function_values()
        all_f = list(set(all_f))
        all_f.sort()

        # Get differences between the function value and all other function values
        diff = np.abs(np.array(all_f) - a)
        diff = diff[diff != 0]
        diff.sort()

        # For each difference, build the slice and see if the two vertices are in the same connected component
        for n in diff:
            S = self.slice(a - n, a + n, type="closed")
            try:
                nx.shortest_path(S.to_undirected(), u, v)
                return n
            except:
                pass

    # -----------------------------------------------#
    # Methods for adding and removing nodes and edges
    # -----------------------------------------------#

    def get_next_vert_name(self):
        """Get the next name for a vertex in the Reeb graph that isn't already included. If there are no nodes, it will return 0. Otherwise, it will name the next vertex as one more than the maximum integer labeled vertex, ignoring any strings.

        Returns:
            str or int
                The next name in the sequence.
        """
        integer_nodes = [v for v in self.nodes.keys() if type(v) == int]
        if len(integer_nodes) == 0:
            return 0
        else:
            return max(integer_nodes) + 1

    def add_node(self, vertex, f_vertex, reset_pos=True):
        """Add a vertex to the Reeb graph.
        If the vertex name is given as None, it will be assigned via the get_next_vert_name method.

        Parameters:
            vertex (hashable like int or str, or None) : The name of the vertex to add.
            f_vertex (float) : The function value of the vertex being added.
            reset_pos (bool, optional)
                If True, will reset the positions of the nodes based on the function values.
        """
        if vertex in self.nodes:
            raise ValueError(f"The vertex {vertex} is already in the Reeb graph.")

        if vertex is None:
            vertex = self.get_next_vert_name()

        super().add_node(vertex)

        self.f[vertex] = f_vertex

        if reset_pos:
            self.set_pos_from_f()

    def add_nodes_from(self, nodes, f_dict, reset_pos=True):
        """Add a list of nodes to the Reeb graph.

        Parameters:
            nodes (list)
                The list of node names to add.
            f_dict (dict): A dictionary of function values associated with the nodes. Should be ``f_dict[node] = f_value``.
            reset_pos (bool, optional): If True, will reset the positions of the nodes based on the function values.
        """
        for node in nodes:
            self.add_node(node, f_dict[node], reset_pos=False)

        for node in nodes:
            self.f[node] = f_dict[node]

        if reset_pos:
            self.set_pos_from_f()

    def remove_node(self, vertex, reset_pos=True):
        """Remove a vertex from the Reeb graph.

        Parameters:
            vertex (hashable like int or str): The name of the vertex to remove.
            reset_pos (bool, optional): If True, will reset the positions of the nodes based on the function values.
        """
        if vertex not in self.nodes:
            raise ValueError(f"The vertex {vertex} is not in the Reeb graph.")
        super().remove_node(vertex)
        del self.f[vertex]

        if reset_pos and hasattr(self, "pos_f"):
            del self.pos_f[vertex]
            self.set_pos_from_f()

    def remove_nodes_from(self, nodes, reset_pos=True):
        """Remove a list of nodes from the Reeb graph.

        Parameters:
            nodes (list): The list of node names to remove.
            reset_pos (bool, optional): If True, will reset the positions of the nodes based on the function values.
        """
        for node in nodes:
            self.remove_node(node, reset_pos=False)

        if reset_pos:
            self.set_pos_from_f()

    def add_edge(self, u, v, reset_pos=True):
        """Add an edge to the Reeb graph. Make sure that the edge points to the vertex with higher function value.

        Note that if the edge added is between two vertices with the same function value, it will collapse the two vertices into one.

        Parameters:
            u,v: The edge to add.
            reset_pos (bool): Optional. If True, will reset the positions of the nodes based on the function values.
        """
        for vertex in [u, v]:
            try:
                self.f[vertex]
            except KeyError:
                raise ValueError(
                    f"The vertex {vertex} must be in the Reeb graph to add an edge between them."
                )

        if self.f[u] < self.f[v]:
            edge = (u, v)
            super().add_edge(*edge)
        elif self.f[u] > self.f[v]:
            edge = (v, u)
            super().add_edge(*edge)
        else:
            # the function values are the same, so the edge collapses the two vertices
            # wlog we're going to get rid of v, and add all its edges to u

            # get the edges of v
            edges_in = self.in_edges(v)
            edges_out = self.out_edges(v)

            # add the edges to u
            for e in edges_in:
                self.add_edge(e[0], u)
            for e in edges_out:
                self.add_edge(u, e[1])

            # Remove v
            self.remove_node(v)

        if reset_pos:
            self.set_pos_from_f()

    def add_edges_from(self, edges, reset_pos=True):
        """Add a list of edges to the Reeb graph.

        Parameters:
            edges (list): The list of edges to add.
            reset_pos (bool): Optional. If True, will reset the positions of the nodes based on the function values.
        """
        for edge in edges:
            self.add_edge(*edge, reset_pos=False)

        if reset_pos:
            self.set_pos_from_f()

    def subdivide_edge(self, u, v, w, f_w):
        """Subdivide an edge with a new vertex.

        Parameters:
            u,v: The edge to subdivide.
            w: The new vertex to add.
            f_w: The function value of the new vertex.
        """
        f_lower = min(self.f[u], self.f[v])
        f_upper = max(self.f[u], self.f[v])
        if f_w < f_lower or f_w > f_upper:
            raise ValueError(
                "The function value of the new vertex must be between the function values of the edge vertices."
            )

        edge = sorted([u, v], key=lambda x: self.f[x])
        if edge not in self.edges:
            raise ValueError("The edge must be in the Reeb graph to subdivide it.")

        self.f[w] = f_w

        self.remove_edge(*edge)
        self.add_node(w, f_w)
        self.add_edge(u, w)
        self.add_edge(w, v)

        self.set_pos_from_f()

    def remove_regular_vertex(self, v):
        """Remove a regular vertex from the Reeb graph. A regular vertex is one for which down degree = up degree = 1, so it can be removed and replaed with a single edge.

        Parameters:
            v (int): The vertex to remove.
        """
        if v not in self.nodes:
            raise ValueError("The vertex must be in the Reeb graph to be removed.")

        if self.up_degree(v) != 1 or self.down_degree(v) != 1:
            raise ValueError(
                "The vertex must have up degree 1 and down degree 1 to be removed as a regular vertex."
            )

        u = list(self.predecessors(v))[0]
        w = list(self.successors(v))[0]

        self.add_edge(u, w)
        self.remove_node(v)

    def remove_all_regular_vertices(self):
        """
        Remove all regular vertices from the Reeb graph.
        """
        regular_vertices = [
            v for v in self.nodes if self.up_degree(v) == 1 and self.down_degree(v) == 1
        ]

        for v in regular_vertices:
            self.remove_regular_vertex(v)

    def remove_isolates(self):
        """
        Remove all isolated vertices from the Reeb graph.
        """
        self.remove_nodes_from(list(nx.isolates(self)))

    # ----------------------------------#
    # Methods for drawing the Reeb graph
    # ----------------------------------#
    def set_pos_from_f(self, seed=None, verbose=False):
        """Set the position of the nodes based on the function values. The result will be the (spring layout x, function value y). Note that this will overwrite the previous positions.

        Parameters:
            verbose (bool): Optional. If True, will print out the function values and the positions.
        """
        if len(self.nodes) == 0:
            if verbose:
                print("Nothing to be done, no nodes here")
            self.pos = {}
            self.pos_f = {}
        else:
            pos = nx.spring_layout(self, seed=seed)
            self.pos = pos

            self.pos_f = {}

            for v in self.nodes:
                self.pos_f[v] = (self.pos[v][0], self.f[v])

            if verbose:
                print("Function values:", self.f)
                print("Positions:", self.pos_f)

    def draw(self, with_labels=True, with_colorbar=False, cpx=0.1, ax=None, **kwargs):
        """
        A drawing of the Reeb graph. Uses the fancy version from cereeberus.compute.draw.

        Parameters:
            with_labels (bool): Optional. If True, will include the labels of the nodes.
            cpx (float): Optional. A parameter that controls "loopiness" of multiedges in the drawing.

        Returns:
            None
        """

        if ax is None:
            ax = plt.gca()

        # This really shouldn't be called ever since this is supposed to be maintained by the class
        if not set(self.pos_f.keys()) == set(self.nodes()):
            print("The positions are not set correctly. Setting them now.")
            self.set_pos_from_f()

        draw.reeb_plot(
            self,
            with_labels=with_labels,
            with_colorbar=with_colorbar,
            cpx=cpx,
            cpy=0,
            ax=ax,
            **kwargs,
        )

    # -----------------------------------------------#
    # Methods for getting portions of the Reeb graph
    # -----------------------------------------------#

    def induced_subgraph(self, nodes):
        """
        Returns the subgraph of the Reeb graph induced by the nodes in the list nodes.

        Parameters:
            nodes (list): The list of nodes to include in the subgraph.

        Returns:
            ReebGraph: The subgraph of the Reeb graph induced by the nodes in the list nodes.
        """
        H = ReebGraph(self.subgraph(nodes), f={v: self.f[v] for v in nodes})

        return H

    def slice(self, a, b, type="open", verbose=False):
        """
        Returns the subgraph of the Reeb graph with image in the open interval (a,b) if `type = 'open'` or in the closed interval [a,b] of `type = 'closed'`.
        This will convert any edges that cross the slice into vertices.

        Parameters:
            a (float): The lower bound of the slice.
            b (float): The upper bound of the slice.
            type (str): Optional. The type of interval used to take the slice. Can be 'open' or 'closed'. Default is 'open'.

        Returns:
            ReebGraph: The subgraph of the Reeb graph with image in (a,b).
        """
        if type == "open":
            v_list = [v for v in self.nodes() if self.f[v] > a and self.f[v] < b]
        elif type == "closed":
            v_list = [v for v in self.nodes() if self.f[v] >= a and self.f[v] <= b]

        # Keep the edges where either endpoint (or both) is in (a,b)
        e_list = [e for e in self.edges() if e[0] in v_list or e[1] in v_list]
        # Include the edges that cover the entire slice.
        # Note this assumes that all edges are ordered twoards teh upper function value
        e_list.extend(
            [e for e in self.edges() if self.f[e[0]] < a and self.f[e[1]] > b]
        )

        # Make a dictionary of counts to deal with multiedges
        e_dict = {e: e_list.count(e) for e in e_list}

        if verbose:
            print("Vertices (v,f(v)):", [(v, self.f[v]) for v in v_list])
            print("Edges:", e_list)
            print("Edge dictionary:", e_dict)

        # Build the subgraph
        H = ReebGraph()

        for v in v_list:
            H.add_node(v, self.f[v])

        for e in e_dict:
            if e[0] in v_list and e[1] in v_list:
                # The edge is entirely in the set, so both vertices are already there
                if verbose:
                    print(f"Adding {e_dict[e]} of edge {e} entirely inside slice:")

                for i in range(e_dict[e]):  # Add an edge for each copy in the list
                    H.add_edge(e[0], e[1])

            elif e[0] not in v_list and e[1] not in v_list:
                # The edge is entirely crossing the slice, so we add two vertices and an edge
                if verbose:
                    print(
                        f"Adding {e_dict[e]} of edge {e} with both endpoints outside slice"
                    )

                for i in range(e_dict[e]):
                    v1 = "-".join([str(v) for v in e]) + "_" + str(i) + "_lower"
                    v2 = "-".join([str(v) for v in e]) + "_" + str(i) + "_upper"
                    H.add_node(v1, a)
                    H.add_node(v2, b)
                    H.add_edge(v1, v2)
            else:
                # One vertex is in the set and one is out.
                # Need to check (for the closed case) that this isn't an edge going up from the top bound or down from the bottom bound
                if (
                    e[0] in v_list
                    and self.f[e[0]] == b
                    or e[1] in v_list
                    and self.f[e[1]] == a
                ):
                    if verbose:
                        print(f"Edge {e} is not in the slice")
                    continue

                if verbose:
                    print(
                        f"adding {e_dict[e]} of edge {e} with one endpoint outside slice"
                    )

                # Make a name for the new vertex to create
                for i in range(e_dict[e]):
                    edge_name = "-".join([str(v) for v in e]) + "_" + str(i)

                    if e[0] in v_list:
                        if verbose:
                            print(
                                f"Adding part of edge {e[0],e[1]} with function values {self.f[e[0]],[e[1]]} and {self.f[e[1]]}"
                            )
                            print(
                                f"This should add edge {e[0]}-{edge_name} with function values {self.f[e[0]]} and {b}"
                            )

                        # The lower edge is in the set, so the other vertex must have
                        # value above the max
                        assert self.f[e[1]] >= b
                        vert_in = e[0]
                        func_val = b

                        # Add a new vertex called edge_name with value b
                        H.add_node(edge_name, func_val)
                        H.add_edge(e[0], edge_name)
                    else:
                        # The higher edge is in the set, so the other vertex must have
                        # value below the min
                        if verbose:
                            print(
                                f"Adding part of edge {e[0],e[1]} with function values {self.f[e[0]]} and {self.f[e[1]]}"
                            )
                            print(
                                f"This should add edge ({edge_name}, {e[1]}) with function values {a} and {self.f[e[1]]}"
                            )

                        assert self.f[e[0]] <= a
                        vert_in = e[1]
                        func_val = a

                        # Add a new vertex called edge_name with value a
                        H.add_node(edge_name, func_val)
                        H.add_edge(edge_name, e[1])
        return H

    def connected_components(self):
        """
        Returns the connected components of the Reeb graph by providing a set of vertices for each connected component. If you want to actually get the subgraphs of ReebGraph `R` returned as Reeb graphs, you can use::

            [R.induced_subgraph(component) for component in R.connected_components()]

        Returns:
            A generator of sets of nodes, one for each component of the Reeb graph.

        """

        return nx.connected_components(self.to_undirected())

    def get_largest_component(self):
        """
        Returns the largest connected component of the Reeb graph as a set of nodes. Note: this is not well defined unless the Reeb graph is minimal.

        Returns:
            ReebGraph: The largest connected component of the Reeb graph based on number of vertices.
        """

        v_list = max(self.connected_components(), key=len)
        return self.induced_subgraph(v_list)

    def get_multi_edges(self):
        """
        Returns a set of the edges that have count higher than 1. Useful for checking for multiedges in the Reeb graph.

        Returns:
            set: A set of edges that have count higher than 1.
        """

        mult_edges = [
            (e[0], e[1]) for e in self.edges if self.number_of_edges(e[0], e[1]) > 1
        ]
        return set(mult_edges)

    # -----------------------------------------------#
    # Methods for getting matrices for the graph
    # -----------------------------------------------#

    def adjacency_matrix(self, symmetrize=True):
        """
        Returns the adjacency matrix of the Reeb graph. Symmetrizes the matrix by default; if symmetrize = False, it will return the directed adjacency matrix.

        Parameters:
            symmetrize (bool): Optional. If True, will return the symmetrized adjacency matrix. If False, will return the directed adjacency matrix.

        Returns:
            numpy.ndarray: The adjacency matrix of the Reeb graph.
        """
        A = nx.to_numpy_array(self)
        if symmetrize:
            A = A + A.T

        return A

    def plot_adjacency_matrix(self, symmetrize=True):
        plt.matshow(self.adjacency_matrix(symmetrize=symmetrize))

        # Add vertices as the row and column labels
        plt.xticks(range(len(self.nodes)), self.nodes, rotation=90)
        plt.yticks(range(len(self.nodes)), self.nodes)

    def boundary_matrix(self, astype="numpy"):
        """
        Creates an boundary matrix for the graph, where :math:`B[v,e] = 1` if vertex :math:`v` is an endpoint of edge :math:`e` and :math:`B[v,e] = 0` otherwise. If astype is `map`, it will return a dictionary with keys as edges and values as lists of vertices that are endpoints of that edge.

        Args:
            astype (str): Optional. The type of output to return. Can be 'numpy', or 'map'. Default is 'numpy'.

        Returns:
            numpy.ndarray: The boundary matrix.
        """

        V = list(self.nodes())
        V.sort(key=lambda x: self.f[x])
        E = list(self.edges())
        E.sort(key=lambda x: self.f[x[0]])
        if astype == "numpy":
            B = np.zeros((len(V), len(E)))

            for j, e in enumerate(E):
                i = V.index(e[0])
                B[i, j] = 1

                i = V.index(e[1])
                B[i, j] = 1

            return B

        elif astype == "map":
            B = {}
            for e in E:
                if not len(e) == 3:
                    e = (e[0], e[1], 0)
                B[e] = [e[0], e[1]]
            return B

        else:
            raise ValueError('astype must be either "numpy" or "map".')

    def plot_boundary_matrix(self):
        plt.matshow(self.boundary_matrix())

        # Add vertices as the row labels
        plt.yticks(range(len(self.nodes)), self.nodes)
        plt.xticks(range(len(self.edges)), self.edges, rotation=90)

    # -----------------------------------------------#
    # Operations on Reeb graph to get new Reeb graph
    # -----------------------------------------------#

    def smoothing_and_maps(self, eps=1, verbose=False):
        """
        Builds the ``eps``-smoothed Reeb graph from the input. One way to define this is for a given Reeb graph :math:`(X,f)`, the smoothed graph is the Reeb graph of the product space :math:`(X \\times [-\\varepsilon, \\varepsilon], f(x) +  t)`. This function also returns the maps (i) from the vertices of the original Reeb graph to the vertices of the smoothed Reeb graph and (ii) from the edges of the original to the edges of the smoothed graph.
        These maps are given as a dictionary with keys as vertices (resp. edges) in the original Reeb graph and values as vertices (resp. edges) in the smoothed Reeb graph.

        Parameters:
            eps (float): The amount of smoothing to apply.
            verbose (bool): Optional. If True, will print out additional information during the smoothing process.

        Returns:
            tuple: ReebGraph, vertex_map, edge_map.
        """

        # If eps is 0, return the original graph
        if eps == 0:
            return (
                self,
                {v: v for v in self.nodes},
                {e: [e] for e in self.edges(keys=True)},
            )

        # Get the list of critical values to place new nodes
        crit_vals = list(set(self.f.values()))
        new_crit_vals = [cv + eps for cv in crit_vals]
        new_crit_vals.extend([cv - eps for cv in crit_vals])

        # Add vertices at the locations of the function values of the inputs
        new_crit_vals.extend(list(self.f.values()))
        new_crit_vals = list(set(new_crit_vals))
        new_crit_vals.sort()

        # get smallest diff between adjacent critical values
        # This delta is used for the shift to figure out edges below
        # the vertices at level cv
        min_diff = min(
            [
                new_crit_vals[i + 1] - new_crit_vals[i]
                for i in range(len(new_crit_vals) - 1)
            ]
        )
        delta = min(min_diff / 2, eps / 2)

        # Create the new Reeb graph
        R_eps = ReebGraph()

        cv_0 = (
            new_crit_vals[0] - eps
        )  # Starting value is arbitrary below the first critical value
        H_0 = self.slice(
            cv_0 - eps, cv_0 + eps, type="closed"
        )  # This should initialize empty graph
        C_0 = list(H_0.connected_components())  # This should give empty list
        comp_to_new_vert_0 = {}

        # This will be a dictionary with keys as vertices in R and values as vertices in R_eps
        map_V = {}
        map_E = {e: [] for e in self.edges(keys=True)}

        for i in range(len(new_crit_vals)):
            cv = new_crit_vals[i]
            H = self.slice(cv - eps, cv + eps, type="closed")
            C = list(H.connected_components())

            # ==== Add vertices at level cv =====

            comp_to_new_vert = {}
            # comp_to_new_vert is a dictionary with keys as indices of connected components in the list C and values as the name of the vertex in R_eps
            for i, c in enumerate(C):
                vert_name = R_eps.get_next_vert_name()
                R_eps.add_node(vert_name, cv, reset_pos=False)
                comp_to_new_vert[i] = vert_name

                # Track the vertex map
                for v in c:
                    if v in self.nodes() and self.f[v] == cv:
                        map_V[v] = vert_name

            # ===== Add edges below the vertices at level cv =====

            # Get the slice slightly below the vertex
            H_edge = self.slice(cv - eps - delta, cv + eps - delta, type="closed")
            C_edge = list(H_edge.connected_components())

            # Strip the upper and lower from the strings to be able to check overlaps
            for conn_comp_list in [C, C_0, C_edge]:
                for j, conn_comp in enumerate(conn_comp_list):
                    conn_comp_new = []
                    for i, v in enumerate(conn_comp):
                        if type(v) == str and (
                            v[-6:] == "_upper" or v[-6:] == "_lower"
                        ):
                            conn_comp_new.append(v[:-6])
                        else:
                            conn_comp_new.append(v)
                    conn_comp_list[j] = set(conn_comp_new)

            # Add edges
            # Track the stuff that needs to be included in the map
            V_inv, E_inv = self.inv_image(cv - delta)

            for i, c in enumerate(C_edge):
                overlap_down = np.array([len(c.intersection(c_0)) for c_0 in C_0])
                lower_vert = [
                    comp_to_new_vert_0[u] for u in np.where(overlap_down > 0)[0]
                ]
                if len(lower_vert) > 1:
                    print(f"{i,c} has multiple lower vertices")
                else:
                    lower_vert = lower_vert[0]

                overlap_up = np.array([len(c.intersection(c_0)) for c_0 in C])
                upper_vert = [comp_to_new_vert[u] for u in np.where(overlap_up > 0)[0]]
                if len(upper_vert) > 1:
                    print(f"{i,c} has multiple upper vertices")
                else:
                    upper_vert = upper_vert[0]

                R_eps.add_edge(lower_vert, upper_vert)

                new_edge_name = (
                    lower_vert,
                    upper_vert,
                    R_eps.number_of_edges(lower_vert, upper_vert) - 1,
                )  # Because we just added the new edge, it should be the last one

                # Right now this assumes an endpoint vertex is in the sliced graph, is it possible for it to be edges only?
                for e in E_inv:
                    if e[0] in c or e[1] in c:
                        map_E[e].append(new_edge_name)

            # Set up for the next round
            cv_0 = cv
            H_0 = H
            C_0 = C
            comp_to_new_vert_0 = comp_to_new_vert

        R_eps.set_pos_from_f()

        return R_eps, map_V, map_E

    def smoothing(self, eps=1):
        """
        Builds the ``eps``-smoothed Reeb graph from the input. One way to define this is for a given Reeb graph :math:`(X,f)`, the smoothed graph is the Reeb graph of the product space :math:`(X \\times [-\\varepsilon, \\varepsilon], f(x) +  t)`.

        Parameters:
            eps (float): The amount of smoothing to apply.

        Returns:
            ReebGraph: The smoothed Reeb graph.
        """
        R_eps, _, _ = self.smoothing_and_maps(eps=eps)
        return R_eps

    # -----------------------------------------------#
    # Methods for converting to other graph types
    # -----------------------------------------------#

    def to_mapper(self, delta=None):
        """
        Convert the Reeb graph to a Mapper graph as long as all function values are integers. Note this is NOT the same as computing the mapper graph of a given Reeb graph as the input topological space.  This will create a new Mapper graph object with the same nodes and edges as the Reeb graph.

        Parameters:
            delta (float): Optional. The delta value to use for the Mapper graph. If None, will use 1.

        Returns:
            MapperGraph: The Mapper graph representation of the Reeb graph.
        """

        # Check that all function values are integers before proceeding
        if not all([isinstance(self.f[v], int) for v in self.f]):
            raise ValueError(
                "Function values must be integers to convert to a Mapper graph."
            )

        from cereeberus.reeb.mapper import MapperGraph

        return MapperGraph(self, f=self.f, delta=delta)



import matplotlib.pyplot as plt
import numpy as np
import pulp  # for ILP optimization

from .labeled_blocks import LabeledBlockMatrix as LBM
from .labeled_blocks import LabeledMatrix as LM


# function to build the phi and psi matrices after the ILP optimization
def build_map_matrices(myAssgn, map_results):
    """
    Function to build the map matrices after the ILP optimization. The end result is a dictionary with keys
    - 'phi_0_V', 'phi_0_E'
    - 'phi_n_V', 'phi_n_E'
    - 'psi_0_V', 'psi_0_E'
    - 'psi_n_V', 'psi_n_E'.
    Each key corresponds to a LabeledBlockMatrix containing the respective map matrices for the given thickening.

    Parameters:
        myAssgn (Assignment): the Assignment object
        map_results (dict): the dictionary containing the results of the ILP optimization from the solve_ilp function

    Returns:
        dict: the dictionary containing the final map matrices
    """
    
    # create a dictionary to store the final matrices 
    final_LBMs = {}

    for thickening in ["0", "n"]:
        for obj_type in ["V", "E"]:
            for map_type in ["phi", "psi"]:

                # Gets the full LBM for the relevant interleaving map
                M = myAssgn.get_interleaving_map(
                    maptype=map_type, key=thickening, obj_type=obj_type
                )

                for i in M.get_all_block_indices():
                    block_shape = M[i].array.shape

                    # Reset to zero
                    M[i].array = np.zeros(block_shape)

                    # Set the values according to the ILP optimization results
                    for a in range(block_shape[0]):
                        for b in range(block_shape[1]):
                            # Get the relevant value from the ILP optimization results
                            if a < 0 or b < 0:
                                print(
                                    f"Warning: a={a}, b={b} for block {i} in {map_type}_{thickening}_{obj_type}."
                                )
                            try:
                                M[i].array[a, b] = pulp.value(
                                    map_results[map_type + "_vars"][i][thickening][
                                        obj_type
                                    ][(a, b)]
                                )
                            except KeyError:
                                # TODO: This makes me nervous, but I think the issue is just that there are empty rows/columns that we want to skip
                                # print(f"KeyError: {map_type+'_vars'} for block {i} in {map_type}_{thickening}_{obj_type}.")
                                continue

                output_key = f"{map_type}_{thickening}_{obj_type}"
                final_LBMs[output_key] = M

    return final_LBMs


##------------------- ILP Optimization -------------------##


def solve_ilp(myAssgn, pulp_solver=None, verbose=False):
    """
    Function to solve the ILP feasibility problem for interleaving maps. The function creates a linear programming problem using the PuLP library and solves it to find feasible interleaving maps.

    Parameters:
        myAssgn (Assignment): the Assignment object containing the interleaving maps and other relevant data
        pulp_solver (pulp.LpSolver): the solver to use for the ILP optimization. If None, the default solver is used.
        verbose (bool): whether to print the optimization status and results

    Returns:
        tuple:
            A tuple ``(final_maps, status_str)`` where ``final_maps`` is a dictionary
            of LabeledBlockMatrices when feasible and ``status_str`` is the PuLP
            status string (for example ``"Optimal"``). If infeasible,
            ``final_maps`` is ``None`` and ``status_str`` contains the solver status.
    """
    # function values
    func_vals = myAssgn.all_func_vals()

    # create the ILP problem
    prob = pulp.LpProblem("Interleave_Optimization_Problem", pulp.LpMinimize)

    # create empty dictionaries to store the decision variables
    phi_vars = {
        block: {
            thickening: {obj_type: {} for obj_type in ["V", "E"]}
            for thickening in ["0", "n"]
        }
        for block in func_vals
    }
    psi_vars = {
        block: {
            thickening: {obj_type: {} for obj_type in ["V", "E"]}
            for thickening in ["0", "n"]
        }
        for block in func_vals
    }

    # create the decision variables (NOTE: these are all the decision variables for all the diagrams)
    for thickening in ["0", "n"]:
        for obj_type in ["V", "E"]:
            for block in (func_vals[:-1] if obj_type == "E" else func_vals):

                # create lp variables for phi
                n_rows = myAssgn.phi(thickening, obj_type)[block].get_array().shape[0]
                n_cols = myAssgn.phi(thickening, obj_type)[block].get_array().shape[1]

                phi_vars[block][thickening][obj_type] = pulp.LpVariable.dicts(
                    "phi_" + thickening + obj_type + "_" + str(block),
                    ((a, b) for a in range(n_rows) for b in range(n_cols)),
                    cat="Binary",
                )

                # # set the initial values
                # for a in range(n_rows):
                #     for b in range(n_cols):
                #         prob += phi_vars[block][thickening][obj_type][(a,b)] == myAssgn.phi(thickening, obj_type)[block].get_array()[a][b]

                # create lp variables for psi
                n_rows = myAssgn.psi(thickening, obj_type)[block].get_array().shape[0]
                n_cols = myAssgn.psi(thickening, obj_type)[block].get_array().shape[1]

                psi_vars[block][thickening][obj_type] = pulp.LpVariable.dicts(
                    "psi_" + thickening + obj_type + "_" + str(block),
                    ((a, b) for a in range(n_rows) for b in range(n_cols)),
                    cat="Binary",
                )

                # # set the initial values
                # for a in range(n_rows):
                #     for b in range(n_cols):
                #         prob += psi_vars[block][thickening][obj_type][(a,b)] == myAssgn.psi(thickening, obj_type)[block].get_array()[a][b]

    # create the other decision variables
    z_vars = {
        block: {
            starting_map: {obj_type: {} for obj_type in ["V", "E"]}
            for starting_map in ["F", "G"]
        }
        for block in func_vals
    }

    map_product_vars = {
        block: {
            starting_map: {obj_type: {} for obj_type in ["V", "E"]}
            for starting_map in ["F", "G"]
        }
        for block in func_vals
    }
    for obj_type in ["V", "E"]:
        for starting_map in ["F", "G"]:
            for block in (func_vals[:-1] if obj_type == "E" else func_vals):

                if starting_map == "F":
                    n_rows_1 = myAssgn.psi("n", obj_type)[block].get_array().shape[0]
                    n_cols_1 = myAssgn.phi("0", obj_type)[block].get_array().shape[1]

                    # set the map product variables
                    map_product_vars[block][starting_map][obj_type] = (
                        pulp.LpVariable.dicts(
                            starting_map + "_" + obj_type + "_" + str(block),
                            ((a, b) for a in range(n_rows_1) for b in range(n_cols_1)),
                            cat="Integer",
                        )
                    )

                    n_rowcol_1 = myAssgn.psi("n", obj_type)[block].get_array().shape[1]

                    # set the z variables
                    z_vars[block][starting_map][obj_type] = pulp.LpVariable.dicts(
                        "z_" + starting_map + "_" + obj_type + "_" + str(block),
                        (
                            (a, b, c)
                            for a in range(n_rows_1)
                            for b in range(n_rowcol_1)
                            for c in range(n_cols_1)
                        ),
                        cat="Binary",
                    )

                else:
                    n_rows_1 = myAssgn.phi("n", obj_type)[block].get_array().shape[0]
                    n_cols_1 = myAssgn.psi("0", obj_type)[block].get_array().shape[1]

                    # set the map product variables
                    map_product_vars[block][starting_map][obj_type] = (
                        pulp.LpVariable.dicts(
                            starting_map + "_" + obj_type + "_" + str(block),
                            ((a, b) for a in range(n_rows_1) for b in range(n_cols_1)),
                            cat="Integer",
                        )
                    )

                    n_rowcol_2 = myAssgn.phi("n", obj_type)[block].get_array().shape[1]

                    z_vars[block][starting_map][obj_type] = pulp.LpVariable.dicts(
                        "z_" + starting_map + "_" + obj_type + "_" + str(block),
                        (
                            (a, b, c)
                            for a in range(n_rows_1)
                            for b in range(n_rowcol_2)
                            for c in range(n_cols_1)
                        ),
                        cat="Binary",
                    )

    # create the constraints
    for block in func_vals:
        for starting_map in ['F', 'G']:
            if block not in myAssgn.all_func_vals(map=starting_map): # If the block is not in the starting map, skip it. This is to account for the graphs with different function values
                continue
            # set the other map based on starting map 
            if starting_map == 'F':
                other_map = 'G'
            else:
                other_map = 'F'

            for up_or_down in ['up', 'down']: # deals with 1 (up, down) and 2 (up, down)
                
                if block == myAssgn.all_func_vals(map=starting_map)[-1]: # skip the last block for this type of diagrams
                    continue
                
                # set the matrices
                if up_or_down == 'up': #NOTE: the change in block indices
                    bou_n = myAssgn.B_up(other_map, 'n')[block].get_array()
                    bou_0 = myAssgn.B_up(starting_map, '0')[block].get_array()
                    if starting_map == 'F':
                        map_V = myAssgn.phi('0', 'V')[block+1].get_array()
                        map_E = myAssgn.phi('0', 'E')[block].get_array()
                        map_V_vars = phi_vars[block+1]['0']['V']
                        map_E_vars = phi_vars[block]['0']['E']
                    else:
                        map_V = myAssgn.psi('0', 'V')[block+1].get_array()
                        map_E = myAssgn.psi('0', 'E')[block].get_array()
                        map_V_vars = psi_vars[block+1]['0']['V']
                        map_E_vars = psi_vars[block]['0']['E']
                else:
                    bou_n = myAssgn.B_down(other_map, 'n')[block].get_array()
                    bou_0 = myAssgn.B_down(starting_map, '0')[block].get_array()
                    if starting_map == 'F':
                        map_V = myAssgn.phi('0', 'V')[block].get_array()
                        map_E = myAssgn.phi('0', 'E')[block].get_array()
                        map_V_vars = phi_vars[block]['0']['V']
                        map_E_vars = phi_vars[block]['0']['E']
                    else:
                        map_V = myAssgn.psi('0', 'V')[block].get_array()
                        map_E = myAssgn.psi('0', 'E')[block].get_array()
                        map_V_vars = psi_vars[block]['0']['V']
                        map_E_vars = psi_vars[block]['0']['E']

                # set the dimensions
                shape_m_mix = map_V.shape[0]
                shape_n_mix = map_V.shape[1]
                shape_o_mix = bou_n.shape[1]
                shape_p_mix = map_E.shape[1]

                # constraint 1: loss is bigger than the absolute value of each matrix elements

                for i in range(shape_m_mix):
                    for k in range(shape_p_mix):
                        # inner difference
                        first_term = pulp.lpSum([map_V_vars[i,j] * bou_0[j][k] for j in range(shape_n_mix)])
                        second_term = pulp.lpSum([bou_n[i][l] * map_E_vars[l,k] for l in range(shape_o_mix)])

                        # total expression
                        mixed_expression = (first_term - second_term)

                        prob += mixed_expression == 0

                # constraint 2: each column sums to 1
                for j in range(shape_n_mix):
                    prob += pulp.lpSum(map_V_vars[h,j] for h in range(shape_m_mix)) == 1   

                for k in range(shape_p_mix):
                    prob += pulp.lpSum(map_E_vars[l, k] for l in range(shape_o_mix)) == 1



            for obj_type in ['V', 'E']: # deals with 3, 4, 5, 6, 7, 8, 9, 10
                if obj_type == 'E' and block == func_vals[-1]: # skip the last block for this type of diagrams. This is because we don't have an edge with the highest function value
                    continue

                # multiply inclusion matrices. Needed for the triangles
                i_n_i_0 = myAssgn.I(starting_map, 'n', obj_type)[block].get_array() @ myAssgn.I(starting_map, '0', obj_type)[block].get_array()

                # write inclusion matrices for easier reference. Needed for the parallelograms
                inc_0_para = myAssgn.I(starting_map, '0', obj_type)[block].get_array()
                inc_n_para = myAssgn.I(other_map, 'n', obj_type)[block].get_array()

                
                # set map matrices for easier reference
                if starting_map == 'F':
                    map_0_para_vars = phi_vars[block]['0'][obj_type]
                    map_n_para_vars = phi_vars[block]['n'][obj_type]

                if starting_map == 'G':
                    map_0_para_vars = psi_vars[block]['0'][obj_type]
                    map_n_para_vars = psi_vars[block]['n'][obj_type]

                # set the dimensions
                shape_m_tri = i_n_i_0.shape[0] # for triangles

                shape_o_para = myAssgn.I(other_map, 'n', obj_type)[block].get_array().shape[1] # for parallelograms

                shape_n_para = myAssgn.I(starting_map, '0', obj_type)[block].get_array().shape[0] # for parallelograms

                if starting_map == 'F':
                    shape_n_tri = myAssgn.psi('n', obj_type)[block].get_array().shape[1] # for triangles
                    shape_o_tri = myAssgn.phi('0', obj_type)[block].get_array().shape[1] # for triangles

                    shape_m_para = myAssgn.phi('n', obj_type)[block].get_array().shape[0] # for parallelograms
                    shape_p_para = myAssgn.phi('0', obj_type)[block].get_array().shape[1] # for parallelograms
                else:
                    shape_n_tri = myAssgn.phi('n', obj_type)[block].get_array().shape[1] # for triangles
                    shape_o_tri = myAssgn.psi('0', obj_type)[block].get_array().shape[1] # for triangles

                    # changed shape_n_para = myAssgn.psi('n', obj_type)[block].get_array().shape[1] # for parallelograms

                    shape_m_para = myAssgn.psi('n', obj_type)[block].get_array().shape[0] # for parallelograms
                    shape_p_para = myAssgn.psi('0', obj_type)[block].get_array().shape[1] # for parallelograms


                


                #  constraint 1: loss is bigger than the absolute value of each matrix elements
            
                # for triangles, enforce commutativity entry-wise
                for h in range(shape_m_tri):
                    for k in range(shape_o_tri):
                        prob += (
                            map_product_vars[block][starting_map][obj_type][h, k]
                            == float(i_n_i_0[h, k])
                        )




                # for parallelograms
                for i in range(shape_m_para):
                    for k in range(shape_p_para):
                        # inner difference
                        first_term = pulp.lpSum([map_n_para_vars[i,j] * inc_0_para[j][k] for j in range(shape_n_para)])
                        second_term = pulp.lpSum([inc_n_para[i][l]  * map_0_para_vars[l,k] for l in range(shape_o_para)])


                        # total expression
                        para_expression = first_term - second_term

                        prob += para_expression == 0


                # constraint 2: map_multiplication and z relation. This is for triangles
                for i in range(shape_m_tri):
                    for k in range(shape_o_tri):
                        prob += map_product_vars[block][starting_map][obj_type][i,k] == pulp.lpSum(z_vars[block][starting_map][obj_type][i,j,k] for j in range(shape_n_tri))

                # constraint 3: z is less than either of the map values and greater than sum of the map values minus 1. This is for triangles
                for i in range(shape_m_tri):
                    for j in range(shape_n_tri):
                        for k in range(shape_o_tri):
                            if starting_map == 'F':
                                prob += z_vars[block][starting_map][obj_type][i,j,k] <= psi_vars[block]['n'][obj_type][i,j]
                                prob += z_vars[block][starting_map][obj_type][i,j,k] <= phi_vars[block]['0'][obj_type][j,k]
                                prob += z_vars[block][starting_map][obj_type][i,j,k] >= psi_vars[block]['n'][obj_type][i,j] + phi_vars[block]['0'][obj_type][j,k] - 1
                            else:
                                prob += z_vars[block][starting_map][obj_type][i,j,k] <= phi_vars[block]['n'][obj_type][i,j]
                                prob += z_vars[block][starting_map][obj_type][i,j,k] <= psi_vars[block]['0'][obj_type][j,k]
                                prob += z_vars[block][starting_map][obj_type][i,j,k] >= phi_vars[block]['n'][obj_type][i,j] + psi_vars[block]['0'][obj_type][j,k] - 1

                # constraint 4: each column sums to 1
                if starting_map == 'F':
                    # for triangles
                    for j in range(shape_n_tri):
                        prob += pulp.lpSum(psi_vars[block]['n'][obj_type][i,j] for i in range(shape_m_tri)) == 1                        
                    for k in range(shape_o_tri):
                        prob += pulp.lpSum(phi_vars[block]['0'][obj_type][j,k] for j in range(shape_n_tri)) == 1

                    # for parallelograms
                    for j in range(shape_n_para):
                        prob += pulp.lpSum(phi_vars[block]['n'][obj_type][i,j] for i in range(shape_m_para)) == 1

                    for k in range(shape_p_para):
                        prob += pulp.lpSum(phi_vars[block]['0'][obj_type][j,k] for j in range(shape_o_para)) == 1

                else:
                    # for triangles
                    for j in range(shape_n_tri):
                        prob += pulp.lpSum(phi_vars[block]['n'][obj_type][i,j] for i in range(shape_m_tri)) == 1

                    for k in range(shape_o_tri):
                        prob += pulp.lpSum(psi_vars[block]['0'][obj_type][j,k] for j in range(shape_n_tri)) == 1

                    # for parallelograms
                    for j in range(shape_n_para):
                        prob += pulp.lpSum(psi_vars[block]['n'][obj_type][i,j] for i in range(shape_m_para)) == 1

                    for k in range(shape_p_para):
                        prob += pulp.lpSum(psi_vars[block]['0'][obj_type][j,k] for j in range(shape_o_para)) == 1

    
    # Set the objective function
    prob += 0

    # solve the problem
    if pulp_solver == 'GUROBI':   
        prob.solve(pulp.GUROBI_CMD(msg=0))
    else:
        prob.solve(pulp.PULP_CBC_CMD(msg=0))

    status_str = pulp.LpStatus[prob.status]

    # prob.solve(pulp.GUROBI_CMD(msg=0))
    if prob.status != 1:
        return None, status_str
    
    # create a dictionary to store the results
    map_results = {'phi_vars': phi_vars, 'psi_vars': psi_vars}

    # make the results a LabeledBlockMatrix
    final_maps = build_map_matrices(myAssgn, map_results)
        
    if verbose:
        print("Status:", pulp.LpStatus[prob.status])
        # prob.writeLP("model.lp")  # Write the model in LP format
    
     
    # return results
    return final_maps, status_str



##------------------- ILP Optimization with Distance Matrix -------------------##

def solve_ilp_dist(myAssgn, pulp_solver = None, verbose=False):

    """
    Function to solve the ILP optimization problem for interleaving maps. The function creates a linear programming problem using the PuLP library and solves it to find the optimal interleaving maps.
    
    Parameters:
        myAssgn (Assignment): the Assignment object containing the interleaving maps and other relevant data
        pulp_solver (pulp.LpSolver): the solver to use for the ILP optimization. If None, the default solver is used.
        verbose (bool): whether to print the optimization status and results
        
    Returns:
        tuple: a tuple containing the final interleaving maps (as a dictionary of LabeledBlockMatrices) and the optimized loss value
    """
    # function values
    func_vals = myAssgn.all_func_vals()

    # create the ILP problem
    prob = pulp.LpProblem("Interleave_Optimization_Problem", pulp.LpMinimize)

    # create empty dictionaries to store the decision variables
    phi_vars = {block : {thickening:{obj_type:{} for obj_type in ['V', 'E']} for thickening in ['0', 'n']} for block in func_vals}
    psi_vars = {block : {thickening:{obj_type:{} for obj_type in ['V', 'E']} for thickening in ['0', 'n']} for block in func_vals}


    # create the decision variables (NOTE: these are all the decision variables for all the diagrams)
    for thickening in ['0', 'n']:
        for obj_type in ['V', 'E']:
            for block in (func_vals[:-1] if obj_type == 'E' else func_vals):

                # create lp variables for phi
                n_rows = myAssgn.phi(thickening, obj_type)[block].get_array().shape[0]
                n_cols = myAssgn.phi(thickening, obj_type)[block].get_array().shape[1]

                phi_vars[block][thickening][obj_type] = pulp.LpVariable.dicts('phi_'+thickening+obj_type+'_'+str(block), ((a, b) for a in range(n_rows) for b in range(n_cols)), cat='Binary')

                # # set the initial values
                # for a in range(n_rows):
                #     for b in range(n_cols):
                #         prob += phi_vars[block][thickening][obj_type][(a,b)] == myAssgn.phi(thickening, obj_type)[block].get_array()[a][b]


                # create lp variables for psi
                n_rows = myAssgn.psi(thickening, obj_type)[block].get_array().shape[0]
                n_cols = myAssgn.psi(thickening, obj_type)[block].get_array().shape[1]

                psi_vars[block][thickening][obj_type] = pulp.LpVariable.dicts('psi_'+thickening+obj_type+'_'+str(block), ((a, b) for a in range(n_rows) for b in range(n_cols)), cat='Binary')

                # # set the initial values
                # for a in range(n_rows):
                #     for b in range(n_cols):
                #         prob += psi_vars[block][thickening][obj_type][(a,b)] == myAssgn.psi(thickening, obj_type)[block].get_array()[a][b]

    # create the other decision variables
    z_vars = {block :  {starting_map: {obj_type: {} for obj_type in ['V', 'E']} for starting_map in ['F', 'G']} for block in func_vals}

    map_product_vars = {block : {starting_map: {obj_type: {} for obj_type in ['V', 'E']} for starting_map in ['F', 'G']} for block in func_vals}
    for obj_type in ['V', 'E']:
        for starting_map in ['F', 'G']:
            for block in (func_vals[:-1] if obj_type == 'E' else func_vals):
                
                if starting_map == 'F':
                    n_rows_1 = myAssgn.psi('n', obj_type)[block].get_array().shape[0]
                    n_cols_1 = myAssgn.phi('0', obj_type)[block].get_array().shape[1]
                    
                    # set the map product variables
                    map_product_vars[block][starting_map][obj_type] = pulp.LpVariable.dicts(starting_map+'_'+obj_type+'_'+str(block), ((a, b) for a in range(n_rows_1) for b in range(n_cols_1)), cat='Integer')

                    n_rowcol_1 = myAssgn.psi('n', obj_type)[block].get_array().shape[1]

                    # set the z variables
                    z_vars[block][starting_map][obj_type] = pulp.LpVariable.dicts('z_'+starting_map+'_'+obj_type+'_'+str(block), ((a, b, c) for a in range(n_rows_1) for b in range(n_rowcol_1) for c in range(n_cols_1)), cat='Binary')
                    
                else:
                    n_rows_1 = myAssgn.phi('n', obj_type)[block].get_array().shape[0]
                    n_cols_1 = myAssgn.psi('0', obj_type)[block].get_array().shape[1]

                    # set the map product variables
                    map_product_vars[block][starting_map][obj_type] = pulp.LpVariable.dicts(starting_map+'_'+obj_type+'_'+str(block), ((a, b) for a in range(n_rows_1) for b in range(n_cols_1)), cat='Integer')

                    n_rowcol_2 = myAssgn.phi('n', obj_type)[block].get_array().shape[1]

                    z_vars[block][starting_map][obj_type] = pulp.LpVariable.dicts('z_'+starting_map+'_'+obj_type+'_'+str(block), ((a, b, c) for a in range(n_rows_1) for b in range(n_rowcol_2) for c in range(n_cols_1)), cat='Binary')

    # decison variables for the triangles (to make the ceiling(expression/2 work)
    aux_vars = {
        block: {
            starting_map: {obj_type: {} for obj_type in ["V", "E"]}
            for starting_map in ["F", "G"]
        }
        for block in func_vals
    }

    for obj_type in ["V", "E"]:
        for starting_map in ["F", "G"]:
            # for block in func_vals:
            for block in (func_vals[:-1] if obj_type == "E" else func_vals):
                if starting_map == "F":
                    shape_m_tri = (
                        myAssgn.D("F", "2n", obj_type)[block].get_array().shape[0]
                    )
                else:
                    shape_m_tri = (
                        myAssgn.D("G", "2n", obj_type)[block].get_array().shape[0]
                    )

                aux_vars[block][starting_map][obj_type] = pulp.LpVariable(
                    "aux_" + starting_map + "_" + obj_type + "_" + str(block),
                    cat="Integer",
                )

    # create the minmax variable
    minmax_var = pulp.LpVariable("minmax_var", cat="Integer")

    # create the constraints
    for block in func_vals:
        for starting_map in ["F", "G"]:
            # set the other map based on starting map
            if starting_map == "F":
                other_map = "G"
            else:
                other_map = "F"

            for up_or_down in [
                "up",
                "down",
            ]:  # deals with 1 (up, down) and 2 (up, down)

                if (
                    block == func_vals[-1]
                ):  # skip the last block for this type of diagrams
                    continue

                # set the matrices
                if up_or_down == "up":  # NOTE: the change in block indices
                    dist_n_other = myAssgn.D(other_map, "n", "V")[block + 1].get_array()
                    bou_n = myAssgn.B_up(other_map, "n")[block].get_array()
                    bou_0 = myAssgn.B_up(starting_map, "0")[block].get_array()
                    if starting_map == "F":
                        map_V = myAssgn.phi("0", "V")[block + 1].get_array()
                        map_E = myAssgn.phi("0", "E")[block].get_array()
                        map_V_vars = phi_vars[block + 1]["0"]["V"]
                        map_E_vars = phi_vars[block]["0"]["E"]
                    else:
                        map_V = myAssgn.psi("0", "V")[block + 1].get_array()
                        map_E = myAssgn.psi("0", "E")[block].get_array()
                        map_V_vars = psi_vars[block + 1]["0"]["V"]
                        map_E_vars = psi_vars[block]["0"]["E"]
                else:
                    dist_n_other = myAssgn.D(other_map, "n", "V")[block].get_array()
                    bou_n = myAssgn.B_down(other_map, "n")[block].get_array()
                    bou_0 = myAssgn.B_down(starting_map, "0")[block].get_array()
                    if starting_map == "F":
                        map_V = myAssgn.phi("0", "V")[block].get_array()
                        map_E = myAssgn.phi("0", "E")[block].get_array()
                        map_V_vars = phi_vars[block]["0"]["V"]
                        map_E_vars = phi_vars[block]["0"]["E"]
                    else:
                        map_V = myAssgn.psi("0", "V")[block].get_array()
                        map_E = myAssgn.psi("0", "E")[block].get_array()
                        map_V_vars = psi_vars[block]["0"]["V"]
                        map_E_vars = psi_vars[block]["0"]["E"]

                # set the dimensions
                shape_m_mix = map_V.shape[0]
                shape_n_mix = map_V.shape[1]
                shape_o_mix = bou_n.shape[1]
                shape_p_mix = map_E.shape[1]

                # constraint 1: loss is bigger than the absolute value of each matrix elements

                for i in range(shape_m_mix):
                    for k in range(shape_p_mix):
                        # inner difference
                        first_term = pulp.lpSum(
                            [map_V_vars[i, j] * bou_0[j][k] for j in range(shape_n_mix)]
                        )
                        second_term = pulp.lpSum(
                            [bou_n[i][l] * map_E_vars[l, k] for l in range(shape_o_mix)]
                        )

                        # total expression
                        mixed_expression = pulp.lpSum(
                            dist_n_other[i][h] * (first_term - second_term)
                            for h in range(shape_m_mix)
                        )

                        prob += minmax_var >= mixed_expression

                # constraint 2: each column sums to 1
                for j in range(shape_n_mix):
                    prob += (
                        pulp.lpSum(map_V_vars[h, j] for h in range(shape_m_mix)) == 1
                    )

                for k in range(shape_p_mix):
                    prob += (
                        pulp.lpSum(map_E_vars[l, k] for l in range(shape_o_mix)) == 1
                    )

            for obj_type in ["V", "E"]:  # deals with 3, 4, 5, 6, 7, 8, 9, 10
                if (
                    obj_type == "E" and block == func_vals[-1]
                ):  # skip the last block for this type of diagrams. This is because we don't have an edge with the highest function value
                    continue

                # multiply inclusion matrices. Needed for the triangles
                i_n_i_0 = (
                    myAssgn.I(starting_map, "n", obj_type)[block].get_array()
                    @ myAssgn.I(starting_map, "0", obj_type)[block].get_array()
                )

                # write inclusion matrices for easier reference. Needed for the parallelograms
                inc_0_para = myAssgn.I(starting_map, "0", obj_type)[block].get_array()
                inc_n_para = myAssgn.I(other_map, "n", obj_type)[block].get_array()

                # write dist matrix for easier reference.
                # for triangles
                dist_2n_starting = myAssgn.D(starting_map, "2n", obj_type)[
                    block
                ].get_array()
                # for parallelograms
                dist_2n_other = myAssgn.D(other_map, "2n", obj_type)[block].get_array()

                # set map matrices for easier reference
                if starting_map == "F":
                    map_0_para_vars = phi_vars[block]["0"][obj_type]
                    map_n_para_vars = phi_vars[block]["n"][obj_type]

                if starting_map == "G":
                    map_0_para_vars = psi_vars[block]["0"][obj_type]
                    map_n_para_vars = psi_vars[block]["n"][obj_type]

                # set the dimensions
                shape_m_tri = i_n_i_0.shape[0] # for triangles
                shape_m_para = dist_2n_other.shape[0] # for parallelograms

                shape_o_para = myAssgn.I(other_map, 'n', obj_type)[block].get_array().shape[1] # for parallelograms


                if starting_map == 'F':
                    shape_n_tri = myAssgn.psi('n', obj_type)[block].get_array().shape[1] # for triangles
                    shape_o_tri = myAssgn.phi('0', obj_type)[block].get_array().shape[1] # for triangles

                    shape_n_para = myAssgn.phi('n', obj_type)[block].get_array().shape[1] # for parallelograms
                    
                    shape_p_para = myAssgn.phi('0', obj_type)[block].get_array().shape[1] # for parallelograms
                else:
                    shape_n_tri = (
                        myAssgn.phi("n", obj_type)[block].get_array().shape[1]
                    )  # for triangles
                    shape_o_tri = (
                        myAssgn.psi("0", obj_type)[block].get_array().shape[1]
                    )  # for triangles

                    shape_n_para = (
                        myAssgn.psi("n", obj_type)[block].get_array().shape[1]
                    )  # for parallelograms
                    shape_p_para = (
                        myAssgn.psi("0", obj_type)[block].get_array().shape[1]
                    )  # for parallelograms

                #  constraint 1: loss is bigger than the absolute value of each matrix elements

                # for triangles
                for h in range(shape_m_tri):

                    tri_expression = pulp.lpSum(
                        dist_2n_starting[i, h]
                        * (
                            i_n_i_0[h, k]
                            - map_product_vars[block][starting_map][obj_type][h, k]
                        )
                        for i in range(shape_m_tri)
                        for k in range(shape_o_tri)
                    )

                    prob += (
                        aux_vars[block][starting_map][obj_type] * 2 >= tri_expression
                    )  # ceiling of half of the expression

                    prob += minmax_var >= aux_vars[block][starting_map][obj_type]

                # for parallelograms
                for i in range(shape_m_para):
                    for k in range(shape_p_para):
                        # inner difference
                        first_term = pulp.lpSum(
                            [
                                map_n_para_vars[i, j] * inc_0_para[j][k]
                                for j in range(shape_n_para)
                            ]
                        )
                        second_term = pulp.lpSum(
                            [
                                inc_n_para[i][l] * map_0_para_vars[l, k]
                                for l in range(shape_o_para)
                            ]
                        )

                        # total expression
                        para_expression = pulp.lpSum(
                            dist_2n_other[i][h] * (first_term - second_term)
                            for h in range(shape_m_para)
                        )

                        prob += minmax_var >= para_expression

                # constraint 2: map_multiplication and z relation. This is for triangles
                for i in range(shape_m_tri):
                    for k in range(shape_o_tri):
                        prob += map_product_vars[block][starting_map][obj_type][
                            i, k
                        ] == pulp.lpSum(
                            z_vars[block][starting_map][obj_type][i, j, k]
                            for j in range(shape_n_tri)
                        )

                # constraint 3: z is less than either of the map values and greater than sum of the map values minus 1. This is for triangles
                for i in range(shape_m_tri):
                    for j in range(shape_n_tri):
                        for k in range(shape_o_tri):
                            if starting_map == "F":
                                prob += (
                                    z_vars[block][starting_map][obj_type][i, j, k]
                                    <= psi_vars[block]["n"][obj_type][i, j]
                                )
                                prob += (
                                    z_vars[block][starting_map][obj_type][i, j, k]
                                    <= phi_vars[block]["0"][obj_type][j, k]
                                )
                                prob += (
                                    z_vars[block][starting_map][obj_type][i, j, k]
                                    >= psi_vars[block]["n"][obj_type][i, j]
                                    + phi_vars[block]["0"][obj_type][j, k]
                                    - 1
                                )
                            else:
                                prob += (
                                    z_vars[block][starting_map][obj_type][i, j, k]
                                    <= phi_vars[block]["n"][obj_type][i, j]
                                )
                                prob += (
                                    z_vars[block][starting_map][obj_type][i, j, k]
                                    <= psi_vars[block]["0"][obj_type][j, k]
                                )
                                prob += (
                                    z_vars[block][starting_map][obj_type][i, j, k]
                                    >= phi_vars[block]["n"][obj_type][i, j]
                                    + psi_vars[block]["0"][obj_type][j, k]
                                    - 1
                                )

                # constraint 4: each column sums to 1
                if starting_map == "F":
                    # for triangles
                    for j in range(shape_n_tri):
                        prob += (
                            pulp.lpSum(
                                psi_vars[block]["n"][obj_type][i, j]
                                for i in range(shape_m_tri)
                            )
                            == 1
                        )
                    for k in range(shape_o_tri):
                        prob += (
                            pulp.lpSum(
                                phi_vars[block]["0"][obj_type][j, k]
                                for j in range(shape_n_tri)
                            )
                            == 1
                        )

                    # for parallelograms
                    for j in range(shape_n_para):
                        prob += (
                            pulp.lpSum(
                                phi_vars[block]["n"][obj_type][i, j]
                                for i in range(shape_m_para)
                            )
                            == 1
                        )

                    for k in range(shape_p_para):
                        prob += (
                            pulp.lpSum(
                                phi_vars[block]["0"][obj_type][j, k]
                                for j in range(shape_o_para)
                            )
                            == 1
                        )

                else:
                    # for triangles
                    for j in range(shape_n_tri):
                        prob += (
                            pulp.lpSum(
                                phi_vars[block]["n"][obj_type][i, j]
                                for i in range(shape_m_tri)
                            )
                            == 1
                        )

                    for k in range(shape_o_tri):
                        prob += (
                            pulp.lpSum(
                                psi_vars[block]["0"][obj_type][j, k]
                                for j in range(shape_n_tri)
                            )
                            == 1
                        )

                    # for parallelograms
                    for j in range(shape_n_para):
                        prob += (
                            pulp.lpSum(
                                psi_vars[block]["n"][obj_type][i, j]
                                for i in range(shape_m_para)
                            )
                            == 1
                        )

                    for k in range(shape_p_para):
                        prob += (
                            pulp.lpSum(
                                psi_vars[block]["0"][obj_type][j, k]
                                for j in range(shape_o_para)
                            )
                            == 1
                        )

    # Set the objective function
    prob += minmax_var

    # solve the problem
    if pulp_solver == "GUROBI":
        prob.solve(pulp.GUROBI_CMD(msg=0))
    else:
        prob.solve(pulp.PULP_CBC_CMD(msg=0))

    # prob.solve(pulp.GUROBI_CMD(msg=0))
    if prob.status != 1:
        raise ValueError(
            "The ILP optimization did not converge. Please check the input data and try again."
        )

    # create a dictionary to store the results
    map_results = {"phi_vars": phi_vars, "psi_vars": psi_vars}

    # make the results a LabeledBlockMatrix
    final_maps = build_map_matrices(myAssgn, map_results)

    if verbose:
        print(f"The optimized loss is: {pulp.value(minmax_var)}")
        print("Status:", pulp.LpStatus[prob.status])
        # prob.writeLP("model.lp")  # Write the model in LP format
    

    # return results
    return final_maps, pulp.value(minmax_var)
    return final_maps, pulp.value(minmax_var)



import unittest

import networkx as nx
from sklearn.cluster import DBSCAN
from sklearn.datasets import make_circles

from cereeberus import MapperGraph, computeMapper, cover


class TestReebClass(unittest.TestCase):
    def test_cover(self):
        #Checks to see if the cover generating function is working
        examplecover = [(-1.125, -0.375), (-0.625, 0.125), (-0.125, 0.625), (0.375, 1.125)]
        testcover = cover(min=-1, max=1, numcovers=4, percentoverlap=.5)
        self.assertEqual(examplecover, testcover)

    def test_computeMapper_trivial(self):
        #checking the simple two point graph (mostly to check if the trivial clustering function works, it has been the problem child of this whole project)
        examplegraph1 = MapperGraph()
        examplegraph1.add_node(0,0)
        examplegraph1.add_node(1,1)
        examplegraph1.add_node(2,2)
        examplegraph1.add_edge(0,1)
        testgraph1 = computeMapper([(0.6, 0), (-0.1, 0.5)], (lambda a : a[0]), [(-1,0),(-0.5,0.5),(0,1)], "trivial")
        check = nx.utils.graphs_equal(examplegraph1, testgraph1)
        self.assertEqual(check, True)

    def test_computeMapper_lens_list(self):
        #checks that a list-valued lens function works
        examplegraph = MapperGraph()
        examplegraph.add_node(0,0)
        examplegraph.add_node(1,1)
        examplegraph.add_node(2,2)
        examplegraph.add_edge(0,1)
        pointcloud = [(0.6, 0), (-0.1, 0.5)]
        lens = [0.6, -0.1]
        testgraph = computeMapper(pointcloud, lens, [(-1,0),(-0.5,0.5),(0,1)], "trivial")
        check = nx.utils.graphs_equal(examplegraph, testgraph)
        self.assertEqual(check, True)
        
    def test_computeMapper_nontrivial(self):
        #checking the default graph for the compute_mapper notebook to see if it remains the same
        data, labels = make_circles(n_samples=500, factor=0.4, noise=0.05, random_state=0)
        testgraph2 = computeMapper(data, (lambda a : a[0]), cover(min=-1, max=1, numcovers=7, percentoverlap=.5), DBSCAN(min_samples=2,eps=0.3).fit)
        examplegraph2 = MapperGraph()
        examplegraph2.add_node(0,0)
        examplegraph2.add_node(1,1)
        examplegraph2.add_node(2,1)
        examplegraph2.add_node(3,1)
        examplegraph2.add_node(4,2)
        examplegraph2.add_node(5,2)
        examplegraph2.add_node(6,2)
        examplegraph2.add_node(7,3)
        examplegraph2.add_node(8,3)
        examplegraph2.add_node(9,3)
        examplegraph2.add_node(10,3)
        examplegraph2.add_node(11,4)
        examplegraph2.add_node(12,4)
        examplegraph2.add_node(13,4)
        examplegraph2.add_node(14,5)
        examplegraph2.add_node(15,5)
        examplegraph2.add_node(16,5)
        examplegraph2.add_node(17,6)
        examplegraph2.add_edge(0,1)
        examplegraph2.add_edge(0,2)
        examplegraph2.add_edge(3,4)
        examplegraph2.add_edge(1,6)
        examplegraph2.add_edge(2,5)
        examplegraph2.add_edge(4,7)
        examplegraph2.add_edge(4,8)
        examplegraph2.add_edge(6,10)
        examplegraph2.add_edge(5,9)
        examplegraph2.add_edge(7,11)
        examplegraph2.add_edge(8,11)
        examplegraph2.add_edge(10,13)
        examplegraph2.add_edge(9,12)
        examplegraph2.add_edge(12,16)
        examplegraph2.add_edge(13,15)
        examplegraph2.add_edge(11,14)
        examplegraph2.add_edge(16,17)
        examplegraph2.add_edge(15,17)
        check = nx.utils.graphs_equal(examplegraph2, testgraph2)
        self.assertEqual(check, True)

if __name__ == '__main__':
    unittest.main()


import unittest
from cereeberus import ReebGraph
from cereeberus.data import ex_graphs as ex_g
from cereeberus.data import ex_reebgraphs as ex_rg

class TestReebClass(unittest.TestCase):

    def check_reeb(self, R):
        # A function to do lots of checks on a given Reeb graph. Just run this thing a ton in any unit test to make sure nothing's screwing up the stuff we think we should be maintaining.

        # Check that the node list and the keys for f are the same 
        self.assertEqual(set(R.nodes), set(R.f.keys()))

        # Check that the node list and the keys for pos_f are the same 
        self.assertEqual(set(R.nodes), set(R.pos_f.keys()))

        # Check that the pos_list always has the f value in the second entry 
        for v in R.nodes:
            self.assertEqual(R.pos_f[v][1], R.f[v])

        # Check that all edges are pointing to the higher function value node
        for edge in R.edges:
            v1, v2 = edge[:2]
            self.assertGreater(R.f[v2], R.f[v1])


    def test_reeb_load(self):
        # This test makes sure you can pass in any of the example graph types and convert it to a Reeb graph.
        example_graphs = []
        for ex_graph in [ex_g.simple_loops, ex_g.torus_graph, ex_g.dancing_man, ex_g.juggling_man]:
            G = ex_graph()
            R = ReebGraph(G)
            example_graphs.append(R)
       
            S = {'nodes': len( G.nodes), 'edges': len(G.edges)}
            self.assertEqual(R.summary(), S)

            self.check_reeb(R)
    
    def test_reeb_add_nodes(self):
        # This test makes sure you can add nodes to a Reeb graph.
        R = ex_rg.simple_loops()
        # Test adding a node without specificying the name 
        R.add_node(None, 3)

        self.assertRaises(ValueError, R.add_node, 0, 0.5)
        R.add_node('chicken', 0.5)
        S = {'nodes': 6, 'edges': 5}
        self.assertEqual(R.summary(), S)
        self.check_reeb(R)

        # adding a bunch of nodes 
        v_list = [i for i in range(5, 10)]
        input_f = [0.5, 0.6, 0.7, 0.8, 0.9]
        input_f_dict = {v : f for v, f in zip(v_list, input_f)}
        R.add_nodes_from(v_list, input_f_dict)

        S = {'nodes': 11, 'edges': 5}
        self.assertEqual(R.summary(), S)


        # General check of the Reeb graph
        self.check_reeb(R)


    def test_add_edge(self):
        # This test makes sure you can add an edge to a Reeb graph.
        R = ex_rg.juggling_man()

        # first in, second not
        self.assertRaises(ValueError, R.add_edge, 0, 15)
        # First not in, second in
        self.assertRaises(ValueError, R.add_edge, 15, 0)
        # Neither in the graph
        self.assertRaises(ValueError, R.add_edge, 15, 'chicken')

        # Ok, now this should work 
        # Add an edge between vertices that don't already have an edge
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edge(0, 7)
        S = {'nodes': n, 'edges': m+1}
        self.assertEqual(R.summary(), S)

        # Add an edge between vertices that already have an edge
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edge(0, 1)
        S = {'nodes': n, 'edges': m+1}
        self.assertEqual(R.summary(), S)

        # Add an edge between vertices that are at the same function value. Should merge the vertices into one
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edge(1,4)
        S = {'nodes': n-1, 'edges': m}
        self.assertEqual(R.summary(), S)


        # Check that you can add a bunch of edges
        # These include edges already there, so should get a multi-edge 
        edge_list = [(7,5), (5,10), (5,10), (1,9), (0,9), (7,0)]
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edges_from(edge_list)
        S = {'nodes': n, 'edges': m+6}
        self.assertEqual(R.summary(), S)


        # General check of the Reeb graph
        self.check_reeb(R)

    def test_subdivide_edge(self):
        # This test makes sure you can subdivide an edge in a Reeb graph.
        R = ex_rg.juggling_man()

        # Pass in an existing edge but the function value isn't right 
        self.assertRaises(ValueError, R.subdivide_edge, 0, 1,'chicken', 0.5)

        # Pass in an edge where vertices exist but the edge doesn't
        self.assertRaises(ValueError, R.subdivide_edge, 1,8,'chicken', 0.5)

        # Ok, now actually do it

        n = len(R.nodes)
        m = len(R.edges)
        R.subdivide_edge(6, 7, 'new_v', 2.5)
        # Nodes and edges should be each increased by exactly one 
        S = {'nodes': n+1, 'edges': m+1}
        self.assertEqual(S, R.summary())

        # General check of the Reeb graph
        self.check_reeb(R)

    def test_remove_regular_vertx(self):
        # This test makes sure you can remove a node from a Reeb graph.
        R = ex_rg.torus(multigraph = False)

        # Error if passing in a vertex that isn't already there
        self.assertRaises(ValueError, R.remove_regular_vertex, 'chicken')

        # Error if passing in a vertex but its not regular
        self.assertRaises(ValueError, R.remove_regular_vertex, 'a')
        self.assertRaises(ValueError, R.remove_regular_vertex, 'b')

        # Ok, now actually do it
        n = len(R.nodes)
        m = len(R.edges)
        R.remove_regular_vertex('e')
        S = {'nodes': n-1, 'edges': m-1}
        self.assertEqual(S, R.summary())

        # General check of the Reeb graph
        self.check_reeb(R)

        # One more time with all the regular vertices
        R = ex_rg.torus(multigraph = False)
        n = len(R.nodes)
        m = len(R.edges)
        R.remove_all_regular_vertices()
        S = {'nodes': n-2, 'edges': n-2}
        self.assertEqual(S, R.summary())

        # General check of the Reeb graph
        self.check_reeb(R)

    def test_remove_isolates(self):
        # This test makes sure you can remove isolated nodes from a Reeb graph.
        R = ex_rg.juggling_man()
        R.remove_isolates()

        list_isolates = []
        for v in R.nodes:
            if R.up_degree(v) == 0 and R.down_degree(v) == 0:
                list_isolates.append(v)
        self.assertEqual(len(list_isolates), 0)

        # General check of the Reeb graph
        self.check_reeb(R)
    
    def test_induced_subgraph(self):
        # This test makes sure you can get an induced subgraph from a Reeb graph.
        R = ex_rg.juggling_man()
        v_list = [7,6,3,10,9,1,0]
        H = R.induced_subgraph(v_list)
        self.assertEqual(len(H.nodes), len(v_list))
        self.assertIsInstance(H, ReebGraph)
        self.check_reeb(H)
    
    def test_slice(self):
        # This test makes sure you can slice a Reeb graph.
        R = ex_rg.juggling_man()
        R.add_edge(7,9)


        # Example chosen so that we have vertices with value on the endpoints (we're assuming open interval so shouldn't be included)
        # We also have at least one edge that completely crosses the interval in question
        H = R.slice( 2,5)

        self.assertEqual(H.number_connected_components(),3 )

        # Example chosen so that we have vertices with value on the endpoints (we're using closed interval so now these should be included)
        H = R.slice( 2,5, type = 'closed')

        self.assertEqual(H.number_connected_components(),2 )

    def test_smoothing(self):
        # This test makes sure you can smooth a Reeb graph.
        R = ex_rg.juggling_man()

        R_eps = R.smoothing(0.1)
        self.assertIsInstance(R_eps, ReebGraph)
        self.check_reeb(R_eps)

        # This makes sure that the smoothing behaves with multiedges 
        R = ex_rg.torus()
        R_eps = R.smoothing(0.1)
        # The Euler characteristic should be 0 for a small smoothing
        self.assertEqual(len(R_eps.nodes) - len(R_eps.edges), 0)

    def test_matrices(self):
        # This test makes sure you can get the adjacency matrix and boundary matrix of a Reeb graph.
        R = ex_rg.juggling_man()
        A = R.adjacency_matrix()
        self.assertEqual(A.shape, (len(R.nodes), len(R.nodes)))

        B = R.boundary_matrix()
        self.assertEqual(B.shape, (len(R.nodes), len(R.edges)))




       


if __name__ == '__main__':
    unittest.main()


__all__ = [
    "ReebGraph",
    "LowerStar",
    "computeReeb",
    "computeMapper",
    "MergeTree",
    "MapperGraph",
    "EmbeddedGraph",
    "Interleave",
    "Assignment",
    "data",
    "dist",
    "reeb",
    "compute",
]

from .reeb.reebgraph import ReebGraph
from .reeb.merge import MergeTree
from .reeb.mapper import MapperGraph
from .reeb.embeddedgraph import EmbeddedGraph
from .distance.interleave import Interleave, Assignment
from .reeb.lowerstar import LowerStar
from .compute.computereeb import computeReeb
from .compute.computemapper import computeMapper
from .compute.computemapper import cover

# Examples
from .data import (
    ex_reebgraphs,
    ex_mergetrees,
    ex_mappergraphs,
    ex_embedgraphs,
    ex_torus,
)



import networkx as nx
import numpy as np
import warnings

from networkx.generators import random_labeled_tree

from cereeberus import MergeTree


def randomMergeTree(n=10, func_randomizer=None, range=[0, 10], seed=None):
    """
    Generates a random tree with n nodes + 1 root.
    If `func_randomizer` is None, the function defined is inverted edge count distance from root node.
    Otherwise, func_randomizer should be an input to type the randomizeMergeFunction function. For example, this could be 'uniform' or 'exp'.
    """
    warnings.filterwarnings("ignore")

    # Generate a random tree
    T = random_labeled_tree(n, seed=seed)

    np.random.seed(seed)
    root = np.random.randint(n)

    MT = MergeTree(T, root, seed=seed)

    if func_randomizer is not None:
        MT = randomizeMergeFunction(MT, type=func_randomizer, range=range, seed=seed)

    return MT


def randomizeMergeFunction(MT, range=[0, 10], type="exp", seed=None):
    """
    Returns a merge tree with the same underlying tree, but randomized function values in the given range.
    """
    np.random.seed(seed)
    f = {"v_inf": np.inf}

    for e in nx.edge_bfs(MT.to_undirected(), "v_inf"):
        u = e[0]
        v = e[1]

        bot_f = range[0]
        if u == "v_inf":
            top_f = range[1]
        else:
            top_f = f[u]

        # Here's the main thing.
        # We pull a random number that's lower than the upper
        # neighbor's funciton value.
        # Doing this with uniform repeatedly makes for a squashed merge tree.
        # The expoenential function is a bit better.
        if type == "uniform":
            f[v] = np.random.uniform(bot_f, top_f)
        elif type == "exp":
            f[v] = top_f - np.random.exponential(0.1 * (top_f - bot_f))
            if f[v] < bot_f:
                f[v] = bot_f

    MT.f = f
    MT.set_pos_from_f()

    return MT



# An example lower star filtration built using gudhi's Simplex Tree

from gudhi import SimplexTree
import numpy as np


class LowerStar(SimplexTree):
    """
    Class to create a simplicial complex that has a lower star filtration, based on Gudhi's SimplexTree.

    Inherits from:
        SimplexTree: A simplex tree structure from the Gudhi library.

    Methods:
        __init__(): Initializes the simplex tree.


    """

    def __init__(self):
        """Initialize class. Due to SimplexTree C++ weirdness, this does nothing but initialize the instance."""

        super().__init__()

    def assign_filtration(self, vertex, value):
        """Sets the filtration value for a given vertex. Vertex can be provided either as an integer, or as a single integer in a list representing the integer label of the vertex.

        Args:
            vertex (Union[int, tuple]): Name of the vertex (an integer or a single-element list).
            value (float): Filtration value to set.
        """

        if isinstance(vertex, int):
            vertex = [vertex]
        elif len(vertex) != 1:
            raise ValueError(
                "Lower star filtration can only by updated by specificying the function value on a vertex. Input -vertex- must be an integer, or a single-element list."
            )

        super().assign_filtration(vertex, value)

        # Set all adjacent simplices to have at least this filtration value
        for simplex, filt in self.get_star(vertex):
            simplex_val = max([self.filtration([v]) for v in simplex])
            super().assign_filtration(simplex, simplex_val)

    def max_filtration(self):
        """Returns the maximum filtration value among all simplices in the torus.

        Returns:
            float: Maximum filtration value.
        """
        max_filt = max(f for _, f in self.get_filtration())
        return max_filt

    def min_filtration(self):
        """Returns the minimum filtration value among all simplices in the torus.

        Returns:
            float: Minimum filtration value.
        """
        min_filt = min(f for _, f in self.get_filtration())
        return min_filt

    def assign_random_values(self, min, max, seed=None):
        """Assign uniform random filtrations sampled in (min, max) for all vertices.

        Args:
            min (float): Minimum value for random function values
            max (float): Maximum value for random function value
            seed (int, optional): Random seed for reproducibility. Default is None.
        """
        if seed is not None:
            rng = np.random.default_rng(seed)
            filtvals = rng.uniform(min, max, self.num_vertices())
        else:
            filtvals = np.random.uniform(min, max, self.num_vertices())
        for v, val in zip(range(self.num_vertices()), filtvals):
            self.assign_filtration(v, val)

    def num_vertices(self):
        """Return the number of vertices (0-simplices) in the simplex tree."""
        return sum(1 for simplex in self.get_simplices() if len(simplex[0]) == 1)

    def iter_vertices(self):
        """Yield the names (tuples) of all vertices (0-simplices) in the simplex tree."""
        for simplex in self.get_simplices():
            if len(simplex[0]) == 1:
                yield simplex[0][0]


if __name__ == "__main__":
    LS = LowerStar()
    LS.insert([0, 1, 2])
    LS.insert([1, 3])
    LS.insert([2, 3])

    LS.assign_random_values(0.0, 1.0)

    for simplex, filt in LS.get_filtration():
        print(f"Simplex: {simplex}, Filtration: {filt}")



import unittest
from cereeberus import ReebGraph, MapperGraph
from cereeberus.data import ex_graphs as ex_g
from cereeberus.data import ex_reebgraphs as ex_rg
from cereeberus.data import ex_mappergraphs as ex_mg
import numpy as np

class TestMapperClass(unittest.TestCase):

    def check_mapper(self, MG):
        # A function to do lots of checks on a given Mapper graph. Just run this thing a ton in any unit test to make sure nothing's screwing up the stuff we think we should be maintaining.
        # This is largely the same as the Reeb graph check, but with some additional checks for the Mapper graph.

        # Check that all funciton values are integers
        self.assertTrue(all([isinstance(MG.f[v], int) for v in MG.nodes]))

        # Check that all edges have adjacent function values
        for edge in MG.edges:
            v1, v2 = edge[:2]
            print (v1, v2)
            print(MG.f[v1], MG.f[v2])
            self.assertTrue(MG.f[v2] - MG.f[v1] == 1)

        # Check that the node list and the keys for f are the same 
        self.assertEqual(set(MG.nodes), set(MG.f.keys()))

        # Check that the node list and the keys for pos_f are the same 
        self.assertEqual(set(MG.nodes), set(MG.pos_f.keys()))

        # Check that the pos_list always has delta*f-value in the second entry 
        for v in MG.nodes:
            self.assertEqual(MG.pos_f[v][1], MG.delta*MG.f[v])

        # Check that all edges are pointing to the higher function value node
        for edge in MG.edges:
            v1, v2 = edge[:2]
            self.assertGreater(MG.f[v2], MG.f[v1])

    def test_mapper_load(self):
        # This test makes sure you can pass in any of the example graph types and convert it to a Reeb graph.
        for ex_graph in [ex_mg.simple_loops, ex_mg.torus, ex_mg.dancing_man, ex_mg.juggling_man]:
            MG = ex_graph()

            self.check_mapper(MG)

    
    def test_mapper_add_nodes(self):
        # This test makes sure you can add nodes to a Reeb graph.
        R = ex_mg.simple_loops()
        self.assertRaises(ValueError, R.add_node, 0, 0.5)
        R.add_node('w', 5)
        self.check_mapper(R)

        # adding a bunch of nodes 
        v_list = [i for i in range(5, 10)]
        input_f = [6,7,8, 9,10,]
        input_f_dict = {v : f for v, f in zip(v_list, input_f)}
        R.add_nodes_from(v_list, input_f_dict)

        S = {'nodes': 10, 'edges': 5}
        self.assertEqual(R.summary(), S)

        # General check of the mapper graph
        self.check_mapper(R)


    def test_add_edge(self):
        # This test makes sure you can add an edge to a Reeb graph.
        R = ex_mg.juggling_man()

        # Ok, now this should work 
        # Add an edge between vertices that don't already have an edge
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edge(0, 7)

        # Add an edge between vertices that already have an edge
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edge(0, 1)

        # Add an edge between vertices that are at the same function value. Should merge the vertices into one
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edge(1,4)

        # Check that you can add a bunch of edges
        # These include edges already there, so should get a multi-edge 
        edge_list = [(7,5), (5,10), (5,10), (1,9), (0,9), (7,0)]
        n = len(R.nodes)
        m = len(R.edges)
        R.add_edges_from(edge_list)

        # General check of the mapper graph
        self.check_mapper(R)

    
    def test_induced_subgraph(self):
        # This test makes sure you can get an induced subgraph from a Reeb graph.
        R = ex_mg.juggling_man()
        v_list = [7,6,3,10,9,1,0]
        H = R.induced_subgraph(v_list)
        self.assertEqual(len(H.nodes), len(v_list))
        self.assertIsInstance(H, ReebGraph)
        self.check_mapper(H)
    
    def test_slice(self):
        # This test makes sure you can slice a Reeb graph.
        R = ex_mg.juggling_man()
        R.add_edge(7,9)


        # Example chosen so that we have vertices with value on the endpoints (we're assuming open interval so shouldn't be included)
        # We also have at least one edge that completely crosses the interval in question
        H = R.slice( 2,6)

        self.assertEqual(H.number_connected_components(),3 )

    def test_dist_matrix(self):
        # This test makes sure you can get the distance matrix from the mapper graph.
        R = ex_mg.juggling_man()
        M = R.thickening_distance_by_level(4)
        self.assertEqual(M[5,6], 1)
        self.assertEqual(M[5,5], 0)

        # These should have infinity entries since there are disconnected components at that level
        M = R.thickening_distance_by_level(5)
        self.assertEqual(M[2,10], np.inf)
        self.assertEqual(M[2,3], 1)

        # Check the whole put together matrix
        M = R.thickening_distance_matrix()
        self.assertEqual(M[5][3,10], np.inf)
       


if __name__ == '__main__':
    unittest.main()


import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import networkx as nx


def dict_to_list(d):
    return list(d.values())


def line_loop_index(R):
    """Determine if edges between two nodes should be lines or loops

    Args:
        R (Reeb Graph): Reeb Graph

    Returns:
        2-element tuple containing

        - **line_index (list)** : list of indices for edges to be drawn as lines
        - **loop_index (list)** : list of indices for edges to be drawn as loops
    """
    edge_list = list(R.edges)
    n = len(R.edges)
    loop_index = []
    line_index = []
    for i in range(0, n):
        if edge_list[i][2] == 1:
            loop_index.append(edge_list.index(edge_list[i][0:2] + (0,)))
            loop_index.append(i)
            line_index.remove(edge_list.index(edge_list[i][0:2] + (0,)))
        else:
            line_index.append(i)

    return (line_index, loop_index)


def slope_intercept(pt0, pt1):
    """Compute slope and intercept to be used in the bezier curve function

    Args:
        pt0 (ordered pair): first point
        pt1 (ordered pair): second point

    Returns:
        2-element tuple containing

        - **m (float)** : slope
        - **b (float)** : intercept
    """
    m = (pt0[1] - pt1[1]) / (pt0[0] - pt1[0])
    b = pt0[1] - m * pt0[0]
    return (m, b)


def bezier_curve(pt0, midpt, pt1):
    """Compute bezier curves for plotting two edges between a single set of nodes

    Args:
        pt0 (ordered pair): first point
        midpt (ordered pair): midpoint for bezier curve to pass through
        pt1 (ordered pair): second point

    Returns:
        points (np array): array of points to be used in plotting
    """

    (x1, y1, x2, y2) = (pt0[0], pt0[1], midpt[0], midpt[1])
    (a1, b1) = slope_intercept(pt0, midpt)
    (a2, b2) = slope_intercept(midpt, pt1)
    points = []

    for i in range(0, 100):
        if x1 == x2:
            continue
        else:
            (a, b) = slope_intercept((x1, y1), (x2, y2))
        x = i * (x2 - x1) / 100 + x1
        y = a * x + b
        points.append((x, y))
        x1 += (midpt[0] - pt0[0]) / 100
        y1 = a1 * x1 + b1
        x2 += (pt1[0] - midpt[0]) / 100
        y2 = a2 * x2 + b2
    return points


def reeb_plot(
    R, with_labels=True, with_colorbar=False, cpx=0.1, cpy=0.1, ax=None, **kwargs
):
    """Main plotting function for the Reeb Graph Class

    Parameters:
        R (Reeb Graph): object of Reeb Graph class
        with_labels (bool): parameter to control whether or not to plot labels
        with_colorbar (bool): parameter to control whether or not to plot colorbar
        cp (float): parameter to control curvature of loops in the plotting function. For vertical Reeb graph, only mess with cpx.

    """
    if ax is None:
        fig, ax = plt.subplots()

    viridis = mpl.colormaps["viridis"].resampled(16)

    n = len(R.nodes)

    edge_list = list(R.edges)
    line_index, loop_index = line_loop_index(R)

    # Some weird plotting to make the colored and labeled nodes work.
    # Taking the list of function values from the pos_f dicationary since the infinite node should already have a position set.
    color_map = [R.pos_f[v][1] for v in R.nodes]
    pathcollection = nx.draw_networkx_nodes(
        R, R.pos_f, node_color=color_map, ax=ax, **kwargs
    )
    if with_labels:
        nx.draw_networkx_labels(R, pos=R.pos_f, font_color="black", ax=ax)
    if with_colorbar:
        plt.colorbar(pathcollection)

    for i in line_index:
        node0 = edge_list[i][0]
        node1 = edge_list[i][1]
        x_pos = (R.pos_f[node0][0], R.pos_f[node1][0])
        y_pos = (R.pos_f[node0][1], R.pos_f[node1][1])
        ax.plot(x_pos, y_pos, color="grey", zorder=0)

    for i in loop_index:
        node0 = edge_list[i][0]
        node1 = edge_list[i][1]
        xmid = (R.pos_f[node0][0] + R.pos_f[node1][0]) / 2
        xmid0 = xmid - cpx * xmid
        xmid1 = xmid + cpx * xmid
        ymid = (R.pos_f[node0][1] + R.pos_f[node1][1]) / 2
        ymid0 = ymid - cpy * ymid
        ymid1 = ymid + cpy * ymid
        curve = bezier_curve(R.pos_f[node0], (xmid0, ymid0), R.pos_f[node1])
        c = np.array(curve)
        ax.plot(c[:, 0], c[:, 1], color="grey", zorder=0)
        curve = bezier_curve(R.pos_f[node0], (xmid1, ymid1), R.pos_f[node1])
        c = np.array(curve)
        ax.plot(c[:, 0], c[:, 1], color="grey", zorder=0)

    ax.tick_params(left=True, bottom=False, labelleft=True, labelbottom=False)



from ..distance.labeled_blocks import LabeledBlockMatrix as LBM
from ..distance.labeled_blocks import LabeledMatrix as LM
from ..compute.unionfind import UnionFind
import numpy as np

from .reebgraph import ReebGraph


class MapperGraph(ReebGraph):
    r"""
    A mapper graph structure. This inherits the properties of the Reeb graph in that it is a graph with a function given on the vertices, but with some additional requirements.

    - The values are integers in some range, [n_low, n_low+1, \cdots, n_high], although we consider the funciton values to be [\delta * n_low, \delta* (n_low+1), \cdots, \delta * n_high] for a stored delta
    - If an edge crosses a value, it has a vertex (so that the inverse image of any integer is only vertices, not interiors of edges)
    - An internal delta is stored so that this can be interpreted as function values [\delta * n_low, \delta* (n_low+1), \cdots, \delta * n_high]
    """

    def __init__(self, G=None, f={}, delta=None, seed=None, verbose=False):

        # Check that $f$ values are only integers
        if not all([isinstance(f[v], int) for v in f]):
            raise ValueError("Function values must be integers.")
        if delta is None:
            self.delta = 1
        else:
            self.delta = delta

        super().__init__(G, f, seed, verbose)

        self.mapperify()

    def add_edge(self, u, v, reset_pos=True):
        """
        Add an edge to the graph. This will also update the internal structure to make sure it satisfies the mapper properties.
        """
        super().add_edge(u, v, reset_pos)
        self.mapperify()

    def mapperify(self):
        """
        Take the internal structure and make sure it satisfies the requirement that all edges have adjacent function values.
        """

        # If we're initializing with nothing, this should pass.
        # Note that if self.n_low is None, then self.n_high and self.delta
        # are both None as well but I am not currently checking that.
        try:
            n_low = min(self.f.values())
            n_high = max(self.f.values())

        except:
            return

        for i in range(n_low, n_high + 1):
            e_list = [e for e in self.edges() if self.f[e[0]] < i and self.f[e[1]] > i]

            for e in e_list:
                w_name = self.get_next_vert_name()
                self.subdivide_edge(*e, w_name, i)

    def add_node(self, vertex, f_vertex, reset_pos=True):
        """
        Same as adding a node in Reeb, but with the additional requirement that the function value is an integer.
        """

        if not isinstance(f_vertex, int):
            raise ValueError("Function values must be integers.")
        return super().add_node(vertex, f_vertex, reset_pos)

    def set_pos_from_f(self, seed=None, verbose=False):
        """
        Same as the Reeb graph function, but we want to draw the vertex locations at delta*function value.
        """
        super().set_pos_from_f(seed, verbose)

        for v in self.nodes():
            self.pos_f[v] = (self.pos_f[v][0], self.delta * self.f[v])

    def induced_subgraph(self, nodes):
        """
        Returns the subgraph of the mapper graph induced by the nodes in the list nodes.

        Parameters:
            nodes (list): The list of nodes to include in the subgraph.

        Returns:
            MapperGraph
        """
        R = super().induced_subgraph(nodes)
        return R.to_mapper(self.delta)

    def smoothing_and_maps(self, n=1):
        """
        Compute the smoothing of a mapper graph as given in todo: Cite the paper. Note that the input :math:`n` parameter is related to the integer function values, not the delta-scaled function values.

        Parameters:
            n (int): The amount of smoothing

        Returns:
            tuple: MapperGraph, vertex_map, edge_map
        """
        if type(n) != int:
            raise ValueError("Smoothing amount must be an integer.")

        M_n, V_map, E_map = super().smoothing_and_maps(n)

        # E_map is a dictonary with output lists of edges, but we should only have one edge to one edge in the mapper graph  case
        # This just strips E_map[key] = [(u,v,0)] to instead be E_map[key] = (u,v,0)
        E_map = {key: E_map[key][0] for key in E_map}

        M_n = M_n.to_mapper(self.delta)

        return M_n, V_map, E_map

    def smoothing(self, n=1):
        """
        Compute the smoothing of a mapper graph as given in todo: Cite the paper. Note that the input :math:`n` parameter is related to the integer function values, not the delta-scaled function values.

        Args:
            n (int, optional): Smoothing amount. Defaults to 1.

        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_
        """
        M_n, _, _ = self.smoothing_and_maps(n)
        return M_n

    # ------------------------------#
    # Functions for computing thickening distance matrix
    # ------------------------------#
    def thickening_distance_by_level(self, level, obj_type="V"):
        """
        Get the thickening distance matrix at a given level. This distance is the amount of thickening needed before the given pair of vertices at that level map to the same connected component.

        Parameters:

            level : int. The level to get the thickening distance matrix at.
            obj_type : str. 'V' or 'E' to get the distance matrix for vertices or edges, respectively. Default is 'V'.

        Returns:
            LabeledMatrix

        """
        # Dict to get list of vertices at a level
        LevelVerts = self.func_to_vertex_dict()

        # Dict to get lists of edges at a level
        LevelEdges = self.func_to_edge_dict()

        # Current level to be checking
        L = level

        # Objects to be checking
        if obj_type == "V":
            # Vertices at the current level
            rowLabels = LevelVerts[L]
        elif obj_type == "E":
            # Edges at the current level
            rowLabels = LevelEdges[L]
        else:
            raise ValueError("Type must be 'V' or 'E'.")

        # If there's only one object, the distance is 0.
        # Return the little matrix block
        if len(rowLabels) == 1:
            D = LM(rows=rowLabels, cols=rowLabels)
            return D

        # Max difference to check
        max_diff = max(L - self.min_f(), self.max_f() - L)

        # Distance matrix for this level
        D = np.zeros(shape=(len(rowLabels), len(rowLabels))) - 1
        D += np.identity(len(rowLabels))
        D = LM(D, rows=rowLabels, cols=rowLabels)

        # Initialize a union find object
        UF = UnionFind(list(self.nodes()))

        # Loop through the levels to add edges, and update teh distance matrix if
        # a pair of objects not already merged.
        for k in range(1, max_diff + 1):

            if obj_type == "V":
                up_level = L + k - 1
                down_level = L - k
            elif obj_type == "E":
                up_level = L + k - 1
                down_level = L - k + 1

            U = []
            # Add edges at each up and down level
            # try/except is to ignore entries without edges
            try:
                # up verts
                U.extend(LevelEdges[up_level])
            except:
                pass

            try:
                # down verts
                U.extend(LevelEdges[down_level])
            except:
                pass

            # Add these edges to the union find object
            for e in U:
                UF.union(e[0], e[1])

            if obj_type == "V":
                for v in rowLabels:
                    for u in rowLabels:
                        if (
                            v != u
                            and D.array[rowLabels.index(v)][rowLabels.index(u)] == -1
                        ):
                            if UF.find(v) == UF.find(u):
                                # print(f"found {v} and {u}")
                                D.array[rowLabels.index(v)][rowLabels.index(u)] = k
                                D.array[rowLabels.index(u)][rowLabels.index(v)] = k
            elif obj_type == "E":
                for i, e in enumerate(rowLabels):
                    for j, f in enumerate(rowLabels):
                        if e != f and D.array[i][j] == -1:
                            if UF.find(e[0]) == UF.find(
                                f[0]
                            ):  # and UF.find(e[1]) == UF.find(f[1]):
                                # print(f"found {e} and {f}")
                                D.array[i][j] = k
                                D.array[j][i] = k

            # Check if there are still any entries of -1
            # If there are none, no need to keep adding edges
            if not np.any(D.array == -1):
                break

        # If there are still -1's, set them to np.inf
        if np.any(D.array == -1):
            D.array[D.array == -1] = np.inf

        return D

    def thickening_distance_matrix(self, obj_type="V"):
        """
        Get the thickening distance matrix for the entire mapper graph. This is a labeled block matrix with rows and columns indexed by vertices, and entries given by the thickening distance between the two vertices.

        Returns:
            LabeledBlockMatrix
        """
        if obj_type == "V":
            V_dict = self.func_to_vertex_dict()
            DistMat = LBM(rows_dict=V_dict, cols_dict=V_dict)
        elif obj_type == "E":
            E_dict = self.func_to_edge_dict()
            DistMat = LBM(rows_dict=E_dict, cols_dict=E_dict)

        for i in DistMat.get_all_block_indices():
            D = self.thickening_distance_by_level(i, obj_type=obj_type)
            DistMat[i] = D

        return DistMat
